@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -23,9 +24,9 @@
 
 #define AUDIO_SINK_INTERFACE "org.bluez.audio.Sink"
 
-struct sink *sink_init(struct device *dev);
-void sink_free(struct device *dev);
-gboolean sink_is_active(struct device *dev);
-avdtp_state_t sink_get_state(struct device *dev);
-gboolean sink_new_stream(struct device *dev, struct avdtp *session,
+struct sink *sink_init(struct audio_device *dev);
+void sink_free(struct audio_device *dev);
+gboolean sink_is_active(struct audio_device *dev);
+avdtp_state_t sink_get_state(struct audio_device *dev);
+gboolean sink_new_stream(struct audio_device *dev, struct avdtp *session,
 				struct avdtp_stream *stream);
