@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2005  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -37,7 +37,6 @@
 #include <getopt.h>
 
 #include <sys/socket.h>
-#include <sys/poll.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
@@ -398,10 +397,11 @@ static struct option main_lopts[] = {
 	{ "msdun",	2, 0, 'X' },
 	{ "activesync",	0, 0, 'a' },
 	{ "mrouter",	1, 0, 'm' },
+	{ "dialup",	0, 0, 'u' },
 	{ 0, 0, 0, 0 }
 };
 
-static char main_sopts[] = "hsc:k:Kr:i:lnp::DQ::AESMP:C::P:Xa";
+static char main_sopts[] = "hsc:k:Kr:i:lnp::DQ::AESMP:C::P:Xam:u";
 
 static char main_help[] = 
 	"Bluetooth LAP (LAN Access over PPP) daemon version " VERSION " \n"
@@ -410,6 +410,7 @@ static char main_help[] =
 	"Options:\n"
 	"\t--show --list -l          Show active LAP connections\n"
 	"\t--listen -s               Listen for LAP connections\n"
+	"\t--dialup -u               Pretend to be a dialup/telephone\n"
 	"\t--connect -c <bdaddr>     Create LAP connection\n"
 	"\t--mrouter -m <bdaddr>     Create mRouter connection\n"
 	"\t--search -Q[duration]     Search and connect\n"
@@ -429,7 +430,7 @@ static char main_help[] =
 	"\t--activesync -a           Enable Microsoft ActiveSync networking\n"
 	"\t--cache -C[valid]         Enable address cache\n";
 
-int main(int argc, char **argv)
+int main(int argc, char *argv[])
 {
 	char *dst = NULL, *src = NULL;
 	struct sigaction sa;
@@ -445,6 +446,7 @@ int main(int argc, char **argv)
 
 		case 's':
 			mode = LISTEN;
+			type = LANACCESS;
 			break;
 
 		case 'c':
@@ -539,6 +541,11 @@ int main(int argc, char **argv)
 			type = MROUTER;
 			break;
 
+		case 'u':
+			mode = LISTEN;
+			type = DIALUP;
+			break;
+
 		case 'h':
 		default:
 			printf(main_help);
@@ -590,18 +597,9 @@ int main(int argc, char **argv)
 	sa.sa_handler = sig_hup;
 	sigaction(SIGHUP, &sa, NULL);
 
-	if (detach) {
-		int fd;
-
-		if (fork()) exit(0);
-
-		/* Direct stdin,stdout,stderr to '/dev/null' */
-		fd = open("/dev/null", O_RDWR);
-		dup2(fd, 0); dup2(fd, 1); dup2(fd, 2);
-		close(fd);
-		
-		setsid();
-		chdir("/");
+	if (detach && daemon(0, 0)) {
+		perror("Can't start daemon");
+		exit(1);
 	}
 
 	openlog("dund", LOG_PID | LOG_NDELAY | LOG_PERROR, LOG_DAEMON);
