@@ -4,7 +4,6 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#include <malloc.h>
 #include <string.h>
 #include <limits.h>
 #include <sys/time.h>
@@ -16,6 +15,9 @@
 #include <sys/mman.h>
 #include <sys/file.h>
 #include <ctype.h>
+#include <dlfcn.h>
+#include <dirent.h>
+#include <signal.h>
 
 #include <gmain.h>
 
@@ -63,6 +65,11 @@ struct _GMainLoop {
 	GMainContext *context;
 };
 
+struct _GDir
+{
+	DIR *dirp;
+};
+
 GIOError g_io_channel_read(GIOChannel *channel, gchar *buf, gsize count,
 				gsize *bytes_read)
 {
@@ -194,6 +201,52 @@ gint g_io_channel_unix_get_fd(GIOChannel *channel)
 	return channel->fd;
 }
 
+static int set_flags(int fd, long flags)
+{
+	long arg;
+
+	arg = fcntl(fd, F_GETFL);
+	if (arg < 0)
+		return -errno;
+
+	/* Return if already set */
+	if ((arg & flags) == flags)
+		return 0;
+
+	arg |= flags;
+	if (fcntl(fd, F_SETFL, arg) < 0)
+		return -errno;
+
+	return 0;
+}
+
+GIOStatus g_io_channel_set_flags(GIOChannel *channel, GIOFlags flags,
+				GError **error)
+{
+	int err, fd;
+	long fd_flags = 0;
+
+	if (!channel || channel->closed)
+		return G_IO_STATUS_ERROR;
+
+	fd = g_io_channel_unix_get_fd(channel);
+
+	if (flags & G_IO_FLAG_APPEND)
+		fd_flags |= O_APPEND;
+	if (flags & G_IO_FLAG_NONBLOCK)
+		fd_flags |= O_NONBLOCK;
+
+	err = set_flags(fd, fd_flags);
+	if (err < 0) {
+		if (error)
+			g_set_error(error, 0, 0, "Unable to set flags: %s",
+					strerror(-err));
+		return G_IO_STATUS_ERROR;
+	}
+
+	return G_IO_STATUS_NORMAL;
+}
+
 struct io_watch {
 	guint id;
 	GIOChannel *channel;
@@ -234,7 +287,7 @@ static gboolean g_io_remove_watch(GMainContext *context, guint id)
 	struct io_watch *w;
 
 	for (l = context->io_watches; l != NULL; l = l->next) {
-		w = l->data;	
+		w = l->data;
 
 		if (w->id != id)
 			continue;
@@ -246,7 +299,7 @@ static gboolean g_io_remove_watch(GMainContext *context, guint id)
 	}
 
 	for (l = context->proc_io_watches; l != NULL; l = l->next) {
-		w = l->data;	
+		w = l->data;
 
 		if (w->id != id)
 			continue;
@@ -561,6 +614,11 @@ guint g_timeout_add(guint interval, GSourceFunc function, gpointer data)
 	return t->id;
 }
 
+guint g_timeout_add_seconds(guint interval, GSourceFunc function, gpointer data)
+{
+	return g_timeout_add(interval, function, data);
+}
+
 guint g_idle_add(GSourceFunc function, gpointer data)
 {
 	return g_timeout_add(1, function, data);
@@ -690,7 +748,7 @@ static gboolean child_watch(GIOChannel *io, GIOCondition cond, gpointer user_dat
 }
 
 static void init_child_pipe(void)
-{  
+{
 	struct sigaction action;
 	GIOChannel *io;
 
@@ -780,7 +838,7 @@ gboolean g_spawn_async(const gchar *working_directory,
 	case -1:
 		g_set_error(error, 0, 0, "fork failed: %s", strerror(errno));
 		return FALSE;
-	case 0:	
+	case 0:
 		exec_child(working_directory, argv, envp, flags,
 				child_setup, user_data);
 		break;
@@ -789,7 +847,7 @@ gboolean g_spawn_async(const gchar *working_directory,
 			*child_pid = pid;
 		return TRUE;
 	}
-		
+
 	/* Never reached */
 	return FALSE;
 }
@@ -850,10 +908,10 @@ gboolean g_source_remove(guint tag)
 
 #define CONTINUATION_CHAR(c, val)				\
 	do {							\
-  		if (((c) & 0xc0) != 0x80) /* 10xxxxxx */	\
-  			goto failed;				\
-  		(val) <<= 6;					\
-  		(val) |= (c) & 0x3f;				\
+		if (((c) & 0xc0) != 0x80) /* 10xxxxxx */	\
+			goto failed;				\
+		(val) <<= 6;					\
+		(val) |= (c) & 0x3f;				\
 	} while (0)
 
 #define INCREMENT_AND_CHECK_MAX(p, i, max_len)					\
@@ -862,7 +920,7 @@ gboolean g_source_remove(guint tag)
 		if ((p)[(i)] == '\0' || ((max_len) >= 0 && (i) >= (max_len)))	\
 			goto failed;						\
 	} while (0)
-				
+
 
 gboolean g_utf8_validate(const gchar *str, gssize max_len, const gchar **end)
 {
@@ -910,7 +968,7 @@ two_remaining:
 
 			if (val < min || !UNICODE_VALID(val))
 				goto failed;
-		} 
+		}
 	}
 
 	if (end)
@@ -1084,19 +1142,19 @@ GSList *g_slist_sort(GSList *list, GCompareFunc cmp_func)
 {
 	GSList *l1, *l2;
 
-	if (!list || !list->next) 
+	if (!list || !list->next)
 		return list;
 
-	l1 = list; 
+	l1 = list;
 	l2 = list->next;
 
 	while ((l2 = l2->next) != NULL) {
-		if ((l2 = l2->next) == NULL) 
+		if ((l2 = l2->next) == NULL)
 			break;
 		l1 = l1->next;
 	}
 
-	l2 = l1->next; 
+	l2 = l1->next;
 	l1->next = NULL;
 
 	return g_slist_sort_merge(g_slist_sort(list, cmp_func),
@@ -1316,6 +1374,189 @@ gchar *g_strdelimit(gchar *string, const gchar *delimiters, gchar new_delim)
 	return string;
 }
 
+gchar *g_strconcat(const gchar *string1, ...)
+{
+	gsize l;
+	va_list args;
+	gchar *s, *concat;
+
+	if (!string1)
+		return NULL;
+
+	l = 1 + strlen(string1);
+	va_start(args, string1);
+	s = va_arg(args, gchar *);
+	while (s) {
+		l += strlen(s);
+		s = va_arg(args, gchar *);
+	}
+	va_end (args);
+
+	concat = g_new(gchar, l);
+	concat[0] = '\0';
+
+	va_start(args, string1);
+	s = va_arg(args, gchar*);
+	while (s) {
+		strcat(concat, s);
+		s = va_arg(args, gchar *);
+	}
+	va_end (args);
+
+	return concat;
+}
+
+gsize g_strlcat(gchar *dest, const gchar *src, gsize dest_size)
+{
+	gchar *d = dest;
+	const gchar *s = src;
+	gsize bytes_left = dest_size;
+	gsize dlength;  /* Logically, MIN(strlen(d), dest_size) */
+
+	if (!d || !s)
+		return 0;
+
+	/* Find the end of dst and adjust bytes left but don't go past end */
+	while (*d != 0 && bytes_left-- != 0)
+		d++;
+	dlength = d - dest;
+	bytes_left = dest_size - dlength;
+
+	if (bytes_left == 0)
+		return dlength + strlen(s);
+
+	while (*s != 0) {
+		if (bytes_left != 1) {
+			*d++ = *s;
+			bytes_left--;
+		}
+		s++;
+	}
+	*d = 0;
+
+	return dlength + (s - src);  /* count does not include NULL */
+}
+
+gchar **g_strsplit(const gchar *string, const gchar *delimiter, gint max_tokens)
+{
+	GSList *string_list = NULL, *slist;
+	gchar **str_array, *s;
+	guint n = 0;
+	const gchar *remainder;
+
+	if (string == NULL || delimiter == NULL || delimiter[0] == '\0')
+		return NULL;
+
+	if (max_tokens < 1)
+		max_tokens = INT_MAX;
+
+	remainder = string;
+	s = strstr(remainder, delimiter);
+	if (s) {
+		gsize delimiter_len = strlen(delimiter);
+
+		while (--max_tokens && s) {
+			gsize len;
+			gchar *tmp;
+
+			len = s - remainder;
+			tmp = g_new(char, len);
+			memcpy(tmp, remainder, len);
+			string_list = g_slist_prepend(string_list, tmp);
+			n++;
+			remainder = s + delimiter_len;
+			s = strstr(remainder, delimiter);
+		}
+	}
+	if (*string) {
+		n++;
+		string_list = g_slist_prepend(string_list, g_strdup(remainder));
+	}
+
+	str_array = g_new(gchar *, n + 1);
+
+	str_array[n--] = NULL;
+	for (slist = string_list; slist; slist = slist->next)
+		str_array[n--] = slist->data;
+
+	g_slist_free(string_list);
+
+	return str_array;
+}
+
+gchar *g_ascii_strup(const gchar *str, gssize len)
+{
+	int i;
+	gchar *s;
+
+	s = g_strdup(str);
+	if (!s)
+		return NULL;
+
+	if (len < 0)
+		len = strlen(s);
+
+	for (i = 0; i < len; i++)
+		s[i] = toupper(s[i]);
+
+	return s;
+}
+
+gboolean g_str_equal(gconstpointer v1, gconstpointer v2)
+{
+	const gchar *string1 = v1;
+	const gchar *string2 = v2;
+
+	return strcmp(string1, string2) == 0;
+}
+
+gboolean g_str_has_prefix(const gchar *str, const gchar *prefix)
+{
+	int str_len;
+	int prefix_len;
+
+	if (str == NULL || prefix == NULL)
+		return FALSE;
+
+	str_len = strlen (str);
+	prefix_len = strlen (prefix);
+
+	if (str_len < prefix_len)
+		return FALSE;
+
+	return strncmp(str, prefix, prefix_len) == 0;
+}
+
+gboolean g_str_has_suffix(const gchar *str, const gchar *suffix)
+{
+	int str_len;
+	int suffix_len;
+
+	if (!str || !suffix)
+		return FALSE;
+
+	str_len = strlen(str);
+	suffix_len = strlen(suffix);
+
+	if (str_len < suffix_len)
+		return FALSE;
+
+	return strcmp(str + str_len - suffix_len, suffix) == 0;
+}
+
+void g_strfreev(gchar **str_array)
+{
+	int i;
+
+	if (str_array == NULL)
+		return;
+
+	for(i = 0; str_array[i] != NULL; i++)
+		g_free(str_array[i]);
+
+	g_free(str_array);
+}
+
 /* GKeyFile */
 
 struct _GKeyFile {
@@ -1364,7 +1605,7 @@ gchar *g_key_file_get_string(GKeyFile *key_file,
 	struct stat st;
 	char *map, *line, *group = NULL, *value = NULL;
 	off_t size;
-	size_t key_len, group_len; 
+	size_t key_len, group_len;
 	int fd, err = 0;
 
 	fd = open(key_file->filename, O_RDONLY);
@@ -1480,6 +1721,56 @@ gboolean g_key_file_get_boolean(GKeyFile *key_file,
 	return ret;
 }
 
+gint g_key_file_get_integer(GKeyFile *key_file,
+				const gchar *group_name,
+				const gchar *key,
+				GError **error)
+{
+	int ret;
+	gchar *str;
+
+	str = g_key_file_get_string(key_file, group_name, key, error);
+	if (!str)
+		return 0;
+
+	ret = atoi(str);
+
+	g_free(str);
+
+	return ret;
+}
+
+gchar **g_key_file_get_string_list(GKeyFile *key_file, const gchar *group_name,
+					const gchar *key, gsize *length,
+					GError **error)
+{
+	gchar *str, *item, **list;
+	int items = 0;
+
+	str = g_key_file_get_string(key_file, group_name, key, error);
+	if (!str)
+		return NULL;
+
+	items = 0;
+	list = g_new0(char *, 1);
+
+	item = strtok(str, ",");
+	while (item) {
+		items++;
+
+		list = g_renew(char *, list, items + 1);
+
+		list[items - 1] = g_strdup(item);
+		list[items] = NULL;
+
+		item = strtok(NULL, ",");
+	}
+
+	g_free(str);
+
+	return list;
+}
+
 /* GString */
 
 #define MY_MAXSIZE ((gsize)-1)
@@ -1511,7 +1802,7 @@ static GString *g_string_sized_new(gsize dfl_size)
 	GString *string;
 
 	string = g_new0(GString, 1);
-	
+
 	g_string_maybe_expand(string, dfl_size);
 	string->str[0] = '\0';
 
@@ -1569,11 +1860,118 @@ gchar *g_string_free(GString *string, gboolean free_segment)
 	if (free_segment) {
 		g_free(string->str);
 		segment = NULL;
-	}
-	else
+	} else
 		segment = string->str;
 
 	g_free(string);
 
 	return segment;
 }
+
+/* GMarkup */
+
+struct _GMarkupParseContext {
+	char dummy;
+};
+
+GMarkupParseContext *g_markup_parse_context_new(const GMarkupParser *parser,
+					GMarkupParseFlags flags,
+					gpointer user_data,
+					GDestroyNotify user_data_dnotify)
+{
+	return g_new0(GMarkupParseContext, 1);
+}
+
+gboolean g_markup_parse_context_parse(GMarkupParseContext *context,
+					const gchar *text, gssize text_len,
+					GError **error)
+{
+	g_set_error(error, 0, 0, "Not implemented");
+	return FALSE;
+}
+
+void g_markup_parse_context_free(GMarkupParseContext *context)
+{
+	g_free(context);
+}
+
+static gchar *g_build_pathname_va(const gchar *first_element,
+						va_list args, gpointer *data)
+{
+	gchar result[PATH_MAX], *element;
+
+	strncpy(result, first_element, PATH_MAX);
+	element = va_arg(args, gchar *);
+
+	while (element) {
+		g_strlcat(result, "/", PATH_MAX);
+		g_strlcat(result, element, PATH_MAX);
+		element = va_arg(args, gchar *);
+	}
+
+	va_end(args);
+
+	return g_strdup(result);
+}
+
+gchar *g_build_filename(const gchar *first_element, ...)
+{
+	gchar *str;
+	va_list args;
+
+	va_start(args, first_element);
+	str = g_build_pathname_va(first_element, args, NULL);
+	va_end(args);
+
+	return str;
+}
+
+/* GDir */
+
+GDir *g_dir_open(const gchar *path, guint flags, GError **error)
+{
+	GDir *dir;
+
+	if (path == NULL)
+		return NULL;
+
+	dir = g_new(GDir, 1);
+
+	dir->dirp = opendir(path);
+
+	if (dir->dirp)
+		return dir;
+
+	/* error case */
+	g_set_error(error, 0, 0, "Error opening directory '%s': %s",
+						path, strerror(errno));
+
+	g_free(dir);
+
+	return NULL;
+}
+
+const gchar *g_dir_read_name(GDir *dir)
+{
+	struct dirent *entry;
+
+	if (dir == NULL)
+		return NULL;
+
+	entry = readdir(dir->dirp);
+
+	while (entry && (strcmp(entry->d_name, ".") == 0 ||
+					strcmp(entry->d_name, "..") == 0))
+		entry = readdir(dir->dirp);
+
+	return entry ? entry->d_name : NULL;
+}
+
+void g_dir_close(GDir *dir)
+{
+	if (dir == NULL)
+		return;
+
+	closedir(dir->dirp);
+	g_free(dir);
+}
