@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *  Copyright (C) 2005-2007  Johan Hedberg <johan.hedberg@nokia.com>
  *
  *
@@ -29,6 +30,8 @@
 #include <stdio.h>
 #include <errno.h>
 #include <stdlib.h>
+#include <limits.h>
+#include <sys/param.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
 
@@ -38,17 +41,15 @@
 #include <bluetooth/sdp.h>
 
 #include <glib.h>
-
 #include <dbus/dbus.h>
+#include <gdbus.h>
 
 #include "adapter.h"
-
-#include "dbus.h"
-#include "dbus-helper.h"
+#include "manager.h"
 #include "hcid.h"
 #include "dbus-common.h"
 #include "dbus-service.h"
-#include "dbus-error.h"
+#include "error.h"
 #include "dbus-security.h"
 #include "dbus-hci.h"
 
@@ -64,6 +65,7 @@ struct passkey_agent {
 	GSList *pending_requests;
 	int exited;
 	guint timeout;
+	guint listener_id;
 };
 
 struct pending_agent_request {
@@ -82,15 +84,17 @@ struct authorization_agent {
 	char *name;
 	char *path;
 	GSList *pending_requests;
+	guint listener_id;
 };
 
 struct auth_agent_req {
-	DBusMessage *msg;
 	struct authorization_agent *agent;
 	char *adapter_path;
 	char *address;
 	char *service_path;
 	char *uuid;
+	service_auth_cb cb;
+	void *user_data;
 	DBusPendingCall *call;
 };
 
@@ -109,10 +113,14 @@ static void passkey_agent_free(struct passkey_agent *agent)
 
 	for (l = agent->pending_requests; l != NULL; l = l->next) {
 		struct pending_agent_request *req = l->data;
+		struct adapter *adapter = manager_find_adapter(&req->sba);
 
 		hci_send_cmd(req->dev, OGF_LINK_CTL,
 				OCF_PIN_CODE_NEG_REPLY, 6, &req->bda);
 
+		if (adapter)
+			adapter_auth_request_replied(adapter, &req->bda);
+
 		send_cancel_request(req);
 	}
 
@@ -134,25 +142,17 @@ static void passkey_agent_free(struct passkey_agent *agent)
 	g_free(agent);
 }
 
-static void agent_exited(const char *name, struct adapter *adapter)
+static void agent_exited(void *user_data)
 {
-	GSList *cur, *next;
+	struct passkey_agent *agent = user_data;
+	struct adapter *adapter = agent->adapter;
 
-	debug("Passkey agent %s exited without calling Unregister", name);
+	debug("Passkey agent exited without calling Unregister");
 
-	for (cur = adapter->passkey_agents; cur != NULL; cur = next) {
-		struct passkey_agent *agent = cur->data;
+	agent->exited = 1;
 
-		next = cur->next;
-
-		if (strcmp(agent->name, name))
-			continue;
-
-		agent->exited = 1;
-
-		adapter->passkey_agents = g_slist_remove(adapter->passkey_agents, agent);
-		passkey_agent_free(agent);
-	}
+	adapter->passkey_agents = g_slist_remove(adapter->passkey_agents, agent);
+	passkey_agent_free(agent);
 }
 
 static gboolean agent_timeout(struct passkey_agent *agent)
@@ -171,15 +171,10 @@ static gboolean agent_timeout(struct passkey_agent *agent)
 	return FALSE;
 }
 
-static void default_agent_exited(const char *name, void *data)
+static void default_agent_exited(void *data)
 {
-	debug("%s exited without unregistering the default passkey agent", name);
-
-	if (!default_agent || strcmp(name, default_agent->name)) {
-		/* This should never happen (there's a bug in the code if it does) */
-		debug("default_agent_exited: mismatch with actual default_agent");
-		return;
-	}
+	debug("D-Bus client exited without unregistering the"
+			" default passkey agent");
 
 	default_agent->exited = 1;
 
@@ -239,17 +234,22 @@ static int agent_cmp(const struct passkey_agent *a, const struct passkey_agent *
 	return 0;
 }
 
-static DBusHandlerResult register_passkey_agent(DBusConnection *conn,
+static inline DBusMessage *invalid_args(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+			"Invalid arguments in method call");
+}
+
+static DBusMessage *register_passkey_agent(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct passkey_agent *agent, ref;
 	struct adapter *adapter;
-	DBusMessage *reply;
 	const char *path, *addr;
 
 	if (!data) {
 		error("register_passkey_agent called without any adapter info!");
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return NULL;
 	}
 
 	adapter = data;
@@ -258,7 +258,10 @@ static DBusHandlerResult register_passkey_agent(DBusConnection *conn,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_STRING, &addr,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
+
+	if ((check_address(addr) < 0) || (path[0] != '/'))
+		return invalid_args(msg);
 
 	memset(&ref, 0, sizeof(ref));
 
@@ -267,44 +270,41 @@ static DBusHandlerResult register_passkey_agent(DBusConnection *conn,
 	ref.path = (char *) path;
 
 	if (g_slist_find_custom(adapter->passkey_agents, &ref, (GCompareFunc) agent_cmp))
-		return error_passkey_agent_already_exists(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".AlreadyExists",
+				"Passkey agent already exists");
 
 	agent = passkey_agent_new(adapter, conn, ref.name, path, addr);
 	if (!agent)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply) {
-		agent->exited = 1;
-		passkey_agent_free(agent);
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-	}
+		return NULL;
 
 	/* Only add a name listener if there isn't one already for this name */
 	ref.addr = NULL;
 	ref.path = NULL;
-	if (!g_slist_find_custom(adapter->passkey_agents, &ref, (GCompareFunc) agent_cmp))
-		name_listener_add(conn, ref.name, (name_cb_t) agent_exited, adapter);
+	if (!g_slist_find_custom(adapter->passkey_agents, &ref,
+						(GCompareFunc) agent_cmp))
+		agent->listener_id = g_dbus_add_disconnect_watch(conn, ref.name,
+						agent_exited, agent, NULL);
 
-	agent->timeout = g_timeout_add(AGENT_TIMEOUT, (GSourceFunc)agent_timeout, agent);
+	agent->timeout = g_timeout_add(AGENT_TIMEOUT,
+					(GSourceFunc) agent_timeout, agent);
 
-	adapter->passkey_agents = g_slist_append(adapter->passkey_agents, agent);
+	adapter->passkey_agents = g_slist_append(adapter->passkey_agents,										agent);
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
 
-static DBusHandlerResult unregister_passkey_agent(DBusConnection *conn,
+static DBusMessage *unregister_passkey_agent(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct adapter *adapter;
 	GSList *match;
 	struct passkey_agent ref, *agent;
-	DBusMessage *reply;
 	const char *path, *addr;
 
 	if (!data) {
 		error("unregister_passkey_agent called without any adapter info!");
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return NULL;
 	}
 
 	adapter = data;
@@ -313,7 +313,7 @@ static DBusHandlerResult unregister_passkey_agent(DBusConnection *conn,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_STRING, &addr,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
 
 	memset(&ref, 0, sizeof(ref));
 
@@ -323,54 +323,50 @@ static DBusHandlerResult unregister_passkey_agent(DBusConnection *conn,
 
 	match = g_slist_find_custom(adapter->passkey_agents, &ref, (GCompareFunc) agent_cmp);
 	if (!match)
-		return error_passkey_agent_does_not_exist(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".DoesNotExist",
+				"Passkey agent does not exist");
 
 	agent = match->data;
 
-	name_listener_remove(agent->conn, agent->name,
-			(name_cb_t) agent_exited, adapter);
+	g_dbus_remove_watch(agent->conn, agent->listener_id);
 
 	adapter->passkey_agents = g_slist_remove(adapter->passkey_agents, agent);
 	agent->exited = 1;
 	passkey_agent_free(agent);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
 
-static DBusHandlerResult register_default_passkey_agent(DBusConnection *conn,
-							DBusMessage *msg, void *data)
+static DBusMessage *register_default_passkey_agent(DBusConnection *conn,
+						DBusMessage *msg, void *data)
 {
-	DBusMessage *reply;
 	const char *path;
 
 	if (default_agent)
-		return error_passkey_agent_already_exists(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".AlreadyExists",
+				"Passkey agent already exists");
 
 	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
 
 	default_agent = passkey_agent_new(NULL, conn, dbus_message_get_sender(msg),
 						path, NULL);
 	if (!default_agent)
 		goto need_memory;
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		goto need_memory;
-
-	name_listener_add(conn, default_agent->name,
-			(name_cb_t) default_agent_exited, NULL);
+	default_agent->listener_id = g_dbus_add_disconnect_watch(conn,
+							default_agent->name,
+							default_agent_exited,
+								NULL, NULL);
 
 	info("Default passkey agent (%s, %s) registered",
 			default_agent->name, default_agent->path);
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 
 need_memory:
 	if (default_agent) {
@@ -379,34 +375,32 @@ need_memory:
 		default_agent = NULL;
 	}
 
-	return DBUS_HANDLER_RESULT_NEED_MEMORY;
+	return NULL;
 }
 
-static DBusHandlerResult unregister_default_passkey_agent(DBusConnection *conn,
+static DBusMessage *unregister_default_passkey_agent(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
-	DBusMessage *reply;
 	const char *path, *name;
 
 	if (!default_agent)
-		return error_passkey_agent_does_not_exist(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".DoesNotExist",
+				"Passkey agent does not exist");
 
 	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
 
 	name = dbus_message_get_sender(msg);
 
 	if (strcmp(name, default_agent->name) || strcmp(path, default_agent->path))
-		return error_passkey_agent_does_not_exist(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".DoesNotExist",
+				"Passkey agent does not exist");
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	name_listener_remove(conn, default_agent->name,
-			(name_cb_t) default_agent_exited, NULL);
+	g_dbus_remove_watch(default_agent->conn, default_agent->listener_id);
 
 	info("Default passkey agent (%s, %s) unregistered",
 			default_agent->name, default_agent->path);
@@ -415,33 +409,34 @@ static DBusHandlerResult unregister_default_passkey_agent(DBusConnection *conn,
 	passkey_agent_free(default_agent);
 	default_agent = NULL;
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
 
-static struct auth_agent_req *auth_agent_req_new(DBusMessage *msg,
-						struct authorization_agent *agent,
+static struct auth_agent_req *auth_agent_req_new(struct authorization_agent *agent,
 						const char *adapter_path,
 						const char *address,
 						const char *service_path,
-						const char *uuid)
+						const char *uuid,
+						service_auth_cb cb,
+						void *user_data)
 {
 	struct auth_agent_req *req;
 
 	req = g_new0(struct auth_agent_req, 1);
 
 	req->agent = agent;
-	req->msg = dbus_message_ref(msg);
 	req->adapter_path = g_strdup(adapter_path);
 	req->address = g_strdup(address);
 	req->service_path = g_strdup(service_path);
 	req->uuid = g_strdup(uuid);
+	req->cb = cb;
+	req->user_data = user_data;
 
 	return req;
 }
 
 static void auth_agent_req_free(struct auth_agent_req *req)
 {
-	dbus_message_unref(req->msg);
 	g_free(req->adapter_path);
 	g_free(req->address);
 	g_free(req->service_path);
@@ -454,8 +449,6 @@ static void auth_agent_req_free(struct auth_agent_req *req)
 static void auth_agent_req_cancel(struct auth_agent_req *req)
 {
 	dbus_pending_call_cancel(req->call);
-	error_canceled(req->agent->conn, req->msg,
-			"Authorization process was canceled");
 }
 
 static void auth_agent_cancel_requests(struct authorization_agent *agent)
@@ -489,7 +482,10 @@ static void auth_agent_call_cancel(struct auth_agent_req *req)
 				DBUS_TYPE_INVALID);
 
 	dbus_message_set_no_reply(message, TRUE);
-	send_message_and_unref(agent->conn, message);
+
+	dbus_connection_send(agent->conn, message, NULL);
+
+	dbus_message_unref(message);
 }
 
 static void auth_agent_free(struct authorization_agent *agent)
@@ -517,17 +513,10 @@ static struct authorization_agent *auth_agent_new(DBusConnection *conn,
 	return agent;
 }
 
-static void default_auth_agent_exited(const char *name, void *data)
+static void default_auth_agent_exited(void *data)
 {
-	debug("%s exited without unregistering the "
-		"default authorization agent", name);
-
-	if (!default_auth_agent || strcmp(name, default_auth_agent->name)) {
-		/* This should never happen! */
-		debug("default_auth_agent_exited: mismatch with "
-			"actual default_auth_agent");
-		return;
-	}
+	debug("D-Bus client exited without unregistering the "
+		"default authorization agent");
 
 	auth_agent_cancel_requests(default_auth_agent);
 	auth_agent_free(default_auth_agent);
@@ -549,44 +538,45 @@ static void auth_agent_release(struct authorization_agent *agent)
 	}
 
 	dbus_message_set_no_reply(message, TRUE);
-	send_message_and_unref(agent->conn, message);
+
+	dbus_connection_send(agent->conn, message, NULL);
+
+	dbus_message_unref(message);
 
 	if (agent == default_auth_agent)
-		name_listener_remove(agent->conn, agent->name,
-				(name_cb_t) default_auth_agent_exited, NULL);
+		g_dbus_remove_watch(agent->conn, agent->listener_id);
 }
 
-static DBusHandlerResult register_default_auth_agent(DBusConnection *conn,
-							DBusMessage *msg,
-							void *data)
+static DBusMessage *register_default_auth_agent(DBusConnection *conn,
+						DBusMessage *msg,
+						void *data)
 {
-	DBusMessage *reply;
 	const char *path;
 
 	if (default_auth_agent)
-		return error_auth_agent_already_exists(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".AlreadyExists",
+				"Authorization agent already exists");
 
 	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
 
 	default_auth_agent = auth_agent_new(conn,
 					dbus_message_get_sender(msg), path);
 	if (!default_auth_agent)
 		goto need_memory;
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		goto need_memory;
-
-	name_listener_add(conn, default_auth_agent->name,
-			(name_cb_t) default_auth_agent_exited, NULL);
+	default_auth_agent->listener_id = g_dbus_add_disconnect_watch(conn,
+						default_auth_agent->name,
+						default_auth_agent_exited,
+								NULL, NULL);
 
 	info("Default authorization agent (%s, %s) registered",
 		default_auth_agent->name, default_auth_agent->path);
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 
 need_memory:
 	if (default_auth_agent) {
@@ -594,36 +584,35 @@ need_memory:
 		default_auth_agent = NULL;
 	}
 
-	return DBUS_HANDLER_RESULT_NEED_MEMORY;
+	return NULL;
 }
 
-static DBusHandlerResult unregister_default_auth_agent(DBusConnection *conn,
+static DBusMessage *unregister_default_auth_agent(DBusConnection *conn,
 							DBusMessage *msg,
 							void *data)
 {
 	const char *path, *name;
-	DBusMessage *reply;
 
 	if (!default_auth_agent)
-		return error_auth_agent_does_not_exist(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".DoesNotExist",
+				"Authorization agent does not exist");
 
 	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &path,
 				DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg);
+		return invalid_args(msg);
 
 	name = dbus_message_get_sender(msg);
 
 	if (strcmp(name, default_auth_agent->name) ||
 		strcmp(path, default_auth_agent->path))
-		return error_auth_agent_does_not_exist(conn, msg);
+		return g_dbus_create_error(msg,
+				ERROR_INTERFACE ".DoesNotExist",
+				"Authorization agent does not exist");
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	name_listener_remove(conn, default_auth_agent->name,
-			(name_cb_t) default_auth_agent_exited, NULL);
+	g_dbus_remove_watch(default_auth_agent->conn,
+				default_auth_agent->listener_id);
 
 	info("Default authorization agent (%s, %s) unregistered",
 		default_auth_agent->name, default_auth_agent->path);
@@ -632,55 +621,23 @@ static DBusHandlerResult unregister_default_auth_agent(DBusConnection *conn,
 	auth_agent_free(default_auth_agent);
 	default_auth_agent = NULL;
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
 
 static void auth_agent_req_reply(DBusPendingCall *call, void *data)
 {
 	struct auth_agent_req *req = data;
-	struct authorization_agent *agent = req->agent;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	DBusMessage *message;
 	DBusError err;
 
 	debug("authorize reply");
 
 	dbus_error_init(&err);
-	if (dbus_set_error_from_message(&err, reply)) {
-		if (dbus_error_has_name(&err, DBUS_ERROR_NO_REPLY))
-			auth_agent_call_cancel(req);
-		error("Authorization agent replied with an error: %s, %s",
-				err.name, err.message);
-		dbus_error_free(&err);
-		goto reject;
-	}
-
-	dbus_error_init(&err);
-	if (!dbus_message_get_args(reply, &err,	DBUS_TYPE_INVALID)) {
-		error("Wrong authorization agent reply signature: %s",
-			err.message);
-		dbus_error_free(&err);
-		goto reject;
-	}
-
-	message = dbus_message_new_method_return(req->msg);
-	if (!message)
-		goto reject;
-
-	send_message_and_unref(agent->conn, message);
-
-	debug("successfull reply was sent");
-
-	goto done;
-
-reject:
-	error_rejected(agent->conn, req->msg);
-
-done:
-	dbus_message_unref(reply);
-
-	agent->pending_requests = g_slist_remove(agent->pending_requests, req);
+	dbus_set_error_from_message(&err, reply);
+	req->cb(&err, req->user_data);
 
+	default_auth_agent->pending_requests =
+		g_slist_remove(default_auth_agent->pending_requests, req);
 	auth_agent_req_free(req);
 
 	debug("auth_agent_reply: returning");
@@ -690,7 +647,7 @@ static DBusPendingCall *auth_agent_call_authorize(struct authorization_agent *ag
 						const char *adapter_path,
 						const char *service_path,
 						const char *address,
-						const char *path)
+						const char *uuid)
 {
 	DBusMessage *message;
 	DBusPendingCall *call;
@@ -706,7 +663,7 @@ static DBusPendingCall *auth_agent_call_authorize(struct authorization_agent *ag
 				DBUS_TYPE_STRING, &adapter_path,
 				DBUS_TYPE_STRING, &address,
 				DBUS_TYPE_STRING, &service_path,
-				DBUS_TYPE_STRING, &path,
+				DBUS_TYPE_STRING, &uuid,
 				DBUS_TYPE_INVALID);
 
 	if (dbus_connection_send_with_reply(agent->conn, message,
@@ -720,44 +677,26 @@ static DBusPendingCall *auth_agent_call_authorize(struct authorization_agent *ag
 	return call;
 }
 
-DBusHandlerResult handle_authorize_request(DBusConnection *conn,
-					DBusMessage *msg,
-					struct service *service,
-					const char *address,
-					const char *uuid)
+int handle_authorize_request_old(struct service *service, const char *path,
+				const char *address, const char *uuid,
+				service_auth_cb cb, void *user_data)
 {
 	struct auth_agent_req *req;
-	char adapter_path[PATH_MAX];
-	bdaddr_t bdaddr;
-	int adapter_id;
-
-	debug("handle_authorize_request");
 
 	if (!default_auth_agent) {
 		debug("no default agent");
-		return error_auth_agent_does_not_exist(conn, msg);
+		return -EPERM;
 	}
 
-	str2ba(address, &bdaddr);
+	req = auth_agent_req_new(default_auth_agent, path,
+				address, service->object_path,
+				uuid, cb, user_data);
 
-	adapter_id = hci_for_each_dev(HCI_UP, find_conn, (long) &bdaddr);
-	if (adapter_id < 0)
-		return error_not_connected(conn, msg);
-
-	debug("Found %s connected to hci%d", address, adapter_id);
-
-	snprintf(adapter_path, sizeof(adapter_path), "/org/bluez/hci%d",
-			adapter_id);
-
-	req = auth_agent_req_new(msg, default_auth_agent, adapter_path,
-					address, service->object_path, uuid);
-
-	req->call = auth_agent_call_authorize(default_auth_agent, adapter_path,
-						service->object_path, address,
-						uuid);
+	req->call = auth_agent_call_authorize(default_auth_agent, path,
+					service->object_path, address, uuid);
 	if (!req->call) {
 		auth_agent_req_free(req);
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return -ENOMEM;
 	}
 
 	dbus_pending_call_set_notify(req->call, auth_agent_req_reply, req,
@@ -767,92 +706,67 @@ DBusHandlerResult handle_authorize_request(DBusConnection *conn,
 
 	debug("authorize request was forwarded");
 
-	return DBUS_HANDLER_RESULT_HANDLED;
+	return 0;
 }
 
-static DBusHandlerResult auth_agent_send_cancel(DBusMessage *msg,
-					struct authorization_agent *agent,
-					const char *adapter_path,
-					struct service *service,
-					const char *address,
-					const char *path)
+static int auth_agent_send_cancel(struct authorization_agent *agent,
+				const char *adapter_path,
+				const char *address)
 {
 	struct auth_agent_req *req = NULL;
-	DBusMessage *message;
 	GSList *l;
 
 	for (l = agent->pending_requests; l != NULL; l = l->next) {
 		req = l->data;
 		if (!strcmp(adapter_path, req->adapter_path) &&
-			!strcmp(address, req->address) &&
-			!strcmp(service->object_path, req->service_path) &&
-			!strcmp(path, req->uuid))
+				!strcmp(address, req->address))
 			break;
 	}
 
 	if (!req)
-		return error_does_not_exist(agent->conn, msg,
-					"No such authorization process");
-
-	message = dbus_message_new_method_return(msg);
-	if (!message)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return -EIO;
 
 	auth_agent_call_cancel(req);
 	auth_agent_req_cancel(req);
 	agent->pending_requests = g_slist_remove(agent->pending_requests, req);
 	auth_agent_req_free(req);
 
-	return send_message_and_unref(agent->conn, message);
+	return 0;
 }
 
-DBusHandlerResult cancel_authorize_request(DBusConnection *conn,
-						DBusMessage *msg,
-						struct service *service,
-						const char *address,
-						const char *path)
+int cancel_authorize_request_old(const char *path, const char *address)
 {
-	char adapter_path[PATH_MAX];
-	int adapter_id;
-	bdaddr_t bdaddr;
-
 	if (!default_auth_agent)
-		return error_auth_agent_does_not_exist(conn, msg);
+		return -EIO;
 
-	str2ba(address, &bdaddr);
-
-	adapter_id = hci_for_each_dev(HCI_UP, find_conn, (long) &bdaddr);
-	if (adapter_id < 0)
-		return error_not_connected(conn, msg);
-
-	snprintf(adapter_path, sizeof(adapter_path), "/org/bluez/hci%d",
-			adapter_id);
-
-	return auth_agent_send_cancel(msg, default_auth_agent, adapter_path,
-						service, address, path);
+	return auth_agent_send_cancel(default_auth_agent, path, address);
 }
 
-static DBusMethodVTable security_methods[] = {
-	{ "RegisterDefaultPasskeyAgent",		register_default_passkey_agent,
-		"s",	""	},
-	{ "UnregisterDefaultPasskeyAgent",		unregister_default_passkey_agent,
-		"s",	""	},
-	{ "RegisterPasskeyAgent",			register_passkey_agent,
-		"ss",	""	},
-	{ "UnregisterPasskeyAgent",			unregister_passkey_agent,
-		"ss",	""	},
-	{ "RegisterDefaultAuthorizationAgent",		register_default_auth_agent,
-		"s",	""	},
-	{ "UnregisterDefaultAuthorizationAgent",	unregister_default_auth_agent,
-		"s",	""	},
-	{ NULL, NULL, NULL, NULL }
+static GDBusMethodTable security_methods[] = {
+	{ "RegisterDefaultPasskeyAgent",	"s",	"",
+					register_default_passkey_agent	},
+	{ "UnregisterDefaultPasskeyAgent",	"s",	"",
+					unregister_default_passkey_agent},
+	{ "RegisterPasskeyAgent",		"ss",	"",
+					register_passkey_agent		},
+	{ "UnregisterPasskeyAgent",		"ss",	"",
+					unregister_passkey_agent	},
+	{ "RegisterDefaultAuthorizationAgent",	"s",	"",
+					register_default_auth_agent	},
+	{ "UnregisterDefaultAuthorizationAgent","s",	"",
+					unregister_default_auth_agent	},
+	{ }
 };
 
 dbus_bool_t security_init(DBusConnection *conn, const char *path)
 {
-	return dbus_connection_register_interface(conn, path, SECURITY_INTERFACE,
-							security_methods,
-							NULL, NULL);
+	return g_dbus_register_interface(conn, path, SECURITY_INTERFACE,
+			security_methods, NULL, NULL, NULL, NULL);
+}
+
+dbus_bool_t security_cleanup(DBusConnection *conn, const char *path)
+{
+	return g_dbus_unregister_interface(conn, path, SECURITY_INTERFACE);
 }
 
 static DBusPendingCall *agent_request(const char *path, bdaddr_t *bda,
@@ -899,6 +813,7 @@ static void passkey_agent_reply(DBusPendingCall *call, void *user_data)
 {
 	struct pending_agent_request *req = user_data;
 	struct passkey_agent *agent = req->agent;
+	struct adapter *adapter = manager_find_adapter(&req->sba);
 	pin_code_reply_cp pr;
 	DBusMessage *message;
 	DBusError err;
@@ -968,6 +883,9 @@ fail:
 			OCF_PIN_CODE_NEG_REPLY, 6, &req->bda);
 
 done:
+	if (adapter)
+		adapter_auth_request_replied(adapter, &req->bda);
+
 	if (message)
 		dbus_message_unref(message);
 
@@ -991,6 +909,7 @@ static int call_passkey_agent(DBusConnection *conn,
 				bdaddr_t *dba)
 {
 	struct pending_agent_request *req;
+	struct adapter *adapter = manager_find_adapter(sba);
 
 	if (!agent) {
 		debug("call_passkey_agent(): no agent available");
@@ -1024,27 +943,19 @@ failed:
 send:
 	hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, dba);
 
+	if (adapter)
+		adapter_auth_request_replied(adapter, dba);
+
 	return -1;
 }
 
-int handle_passkey_request(DBusConnection *conn, int dev, const char *path,
-					bdaddr_t *sba, bdaddr_t *dba)
+int handle_passkey_request_old(DBusConnection *conn, int dev,
+						struct adapter *adapter,
+						bdaddr_t *sba, bdaddr_t *dba)
 {
 	struct passkey_agent *agent = default_agent;
-	struct adapter *adapter = NULL;
 	GSList *l;
 	char addr[18];
-	void *data;
-
-	dbus_connection_get_object_user_data(conn, path, &data);
-
-	if (!data)
-		goto done;
-
-	adapter = data;
-
-	if (!bacmp(&adapter->agents_disabled, dba))
-		goto done;
 
 	ba2str(dba, addr);
 
@@ -1058,187 +969,7 @@ int handle_passkey_request(DBusConnection *conn, int dev, const char *path,
 		}
 	}
 
-done:
-	return call_passkey_agent(conn, agent, dev, path, sba, dba);
-}
-
-static DBusPendingCall *agent_confirm(const char *path, bdaddr_t *bda,
-					struct passkey_agent *agent,
-					const char *value)
-{
-	DBusMessage *message;
-	DBusPendingCall *call;
-	char bda_str[18], *ptr = bda_str;
-
-	message = dbus_message_new_method_call(agent->name, agent->path,
-					"org.bluez.PasskeyAgent", "Confirm");
-	if (message == NULL) {
-		error("Couldn't allocate D-Bus message");
-		return NULL;
-	}
-
-	ba2str(bda, bda_str);
-
-	dbus_message_append_args(message,
-				DBUS_TYPE_STRING, &path,
-				DBUS_TYPE_STRING, &ptr,
-				DBUS_TYPE_STRING, &value,
-				DBUS_TYPE_INVALID);
-
-	if (dbus_connection_send_with_reply(agent->conn, message,
-					&call, REQUEST_TIMEOUT) == FALSE) {
-		error("D-Bus send failed");
-		dbus_message_unref(message);
-		return NULL;
-	}
-
-	dbus_message_unref(message);
-	return call;
-}
-
-static void confirm_agent_reply(DBusPendingCall *call, void *user_data)
-{
-	struct pending_agent_request *req = user_data;
-	struct passkey_agent *agent = req->agent;
-	pin_code_reply_cp pr;
-	DBusMessage *message;
-	DBusError err;
-	int len;
-
-	/* steal_reply will always return non-NULL since the callback
-	 * is only called after a reply has been received */
-	message = dbus_pending_call_steal_reply(call);
-
-	dbus_error_init(&err);
-	if (dbus_set_error_from_message(&err, message)) {
-
-		error("Passkey agent replied with an error: %s, %s",
-				err.name, err.message);
-
-		dbus_error_free(&err);
-		goto fail;
-	}
-
-	dbus_error_init(&err);
-	if (!dbus_message_get_args(message, &err, DBUS_TYPE_INVALID)) {
-		error("Wrong confirm reply signature: %s", err.message);
-		dbus_error_free(&err);
-		goto fail;
-	}
-
-	len = strlen(req->pin);
-
-	set_pin_length(&req->sba, len);
-
-	memset(&pr, 0, sizeof(pr));
-	bacpy(&pr.bdaddr, &req->bda);
-	memcpy(pr.pin_code, req->pin, len);
-	pr.pin_len = len;
-	hci_send_cmd(req->dev, OGF_LINK_CTL,
-			OCF_PIN_CODE_REPLY, PIN_CODE_REPLY_CP_SIZE, &pr);
-
-	goto done;
-
-fail:
-	hci_send_cmd(req->dev, OGF_LINK_CTL,
-			OCF_PIN_CODE_NEG_REPLY, 6, &req->bda);
-
-done:
-	if (message)
-		dbus_message_unref(message);
-
-	agent->pending_requests = g_slist_remove(agent->pending_requests, req);
-	dbus_pending_call_cancel(req->call);
-	if (req->call)
-		dbus_pending_call_unref(req->call);
-	g_free(req->pin);
-	g_free(req->path);
-	g_free(req);
-
-	if (agent != default_agent) {
-		agent->adapter->passkey_agents = g_slist_remove(agent->adapter->passkey_agents,
-								agent);
-		passkey_agent_free(agent);
-	}
-}
-
-static int call_confirm_agent(DBusConnection *conn,
-				struct passkey_agent *agent, int dev,
-				const char *path, bdaddr_t *sba,
-				bdaddr_t *dba, const char *pin)
-{
-	struct pending_agent_request *req;
-
-	if (!agent) {
-		debug("call_passkey_agent(): no agent available");
-		goto send;
-	}
-
-	debug("Calling PasskeyAgent.Confirm: name=%s, path=%s",
-						agent->name, agent->path);
-
-	req = g_new0(struct pending_agent_request, 1);
-	req->dev = dev;
-	bacpy(&req->sba, sba);
-	bacpy(&req->bda, dba);
-	req->agent = agent;
-	req->path = g_strdup(path);
-	req->pin = g_strdup(pin);
-
-	req->call = agent_confirm(path, dba, agent, pin);
-	if (!req->call)
-		goto failed;
-
-	dbus_pending_call_set_notify(req->call, confirm_agent_reply, req, NULL);
-
-	agent->pending_requests = g_slist_append(agent->pending_requests, req);
-
-	return 0;
-
-failed:
-	g_free(req->pin);
-	g_free(req->path);
-	g_free(req);
-
-send:
-	hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, dba);
-
-	return -1;
-}
-
-int handle_confirm_request(DBusConnection *conn, int dev, const char *path,
-				bdaddr_t *sba, bdaddr_t *dba, const char *pin)
-{
-	struct passkey_agent *agent = default_agent;
-	struct adapter *adapter = NULL;
-	GSList *l;
-	char addr[18];
-	void *data;
-
-	dbus_connection_get_object_user_data(conn, path, &data);
-
-	if (!data)
-		goto done;
-
-	adapter = data;
-
-	if (!bacmp(&adapter->agents_disabled, dba))
-		goto done;
-
-	ba2str(dba, addr);
-
-	for (l = adapter->passkey_agents; l != NULL; l = l->next) {
-		struct passkey_agent *a = l->data;
-		if (a != default_agent && g_slist_length(a->pending_requests) >= 1)
-			continue;
-		if (!strcmp(a->addr, addr)) {
-			agent = a;
-			break;
-		}
-	}
-
-done:
-	return call_confirm_agent(conn, agent, dev, path, sba, dba, pin);
+	return call_passkey_agent(conn, agent, dev, adapter->path, sba, dba);
 }
 
 static void send_cancel_request(struct pending_agent_request *req)
@@ -1262,7 +993,9 @@ static void send_cancel_request(struct pending_agent_request *req)
 
 	dbus_message_set_no_reply(message, TRUE);
 
-	send_message_and_unref(req->agent->conn, message);
+	dbus_connection_send(req->agent->conn, message, NULL);
+
+	dbus_message_unref(message);
 
 	debug("PasskeyAgent.Request(%s, %s) was canceled", req->path, address);
 
@@ -1288,24 +1021,26 @@ static void release_agent(struct passkey_agent *agent)
 
 	dbus_message_set_no_reply(message, TRUE);
 
-	send_message_and_unref(agent->conn, message);
+	dbus_connection_send(agent->conn, message, NULL);
+
+	dbus_message_unref(message);
 
 	if (agent == default_agent)
-		name_listener_remove(agent->conn, agent->name,
-				(name_cb_t) default_agent_exited, NULL);
+		g_dbus_remove_watch(agent->conn, agent->listener_id);
 	else {
 		struct passkey_agent ref;
 
-		/* Only remove the name listener if there are no more agents for this name */
+		/* Only remove the name listener if there are no more agents
+		 * for this name */
 		memset(&ref, 0, sizeof(ref));
 		ref.name = agent->name;
-		if (!g_slist_find_custom(agent->adapter->passkey_agents, &ref, (GCompareFunc) agent_cmp))
-			name_listener_remove(agent->conn, ref.name,
-					(name_cb_t) agent_exited, agent->adapter);
+		if (!g_slist_find_custom(agent->adapter->passkey_agents, &ref,
+						(GCompareFunc) agent_cmp))
+			g_dbus_remove_watch(agent->conn, agent->listener_id);
 	}
 }
 
-void release_default_agent(void)
+void release_default_agent_old(void)
 {
 	if (!default_agent)
 		return;
@@ -1333,7 +1068,7 @@ void release_passkey_agents(struct adapter *adapter, bdaddr_t *bda)
 	for (l = adapter->passkey_agents; l != NULL; l = next) {
 		struct passkey_agent *agent = l->data;
 		next = l->next;
-		
+
 		if (bda && agent->addr) {
 			bdaddr_t tmp;
 			str2ba(agent->addr, &tmp);
