@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2000-2001  Qualcomm Incorporated
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -110,7 +110,7 @@ static void hci_req_queue_process(int dev_id)
 
 		data = l->data;
 		data->status = REQ_SENT;
-		
+
 		ret_val = hci_send_cmd(dd, data->ogf, data->ocf, data->clen, data->cparam);
 		if (ret_val < 0) {
 			hci_req_queue = g_slist_remove(hci_req_queue, data);
@@ -187,6 +187,38 @@ static void check_pending_hci_req(int dev_id, int event)
 	hci_req_queue_process(dev_id);
 }
 
+static int get_handle(int dev, bdaddr_t *sba, bdaddr_t *dba, uint16_t *handle)
+{
+	struct hci_conn_list_req *cl;
+	struct hci_conn_info *ci;
+	char addr[18];
+	int i;
+
+	cl = g_malloc0(10 * sizeof(*ci) + sizeof(*cl));
+
+	ba2str(sba, addr);
+	cl->dev_id = hci_devid(addr);
+	cl->conn_num = 10;
+	ci = cl->conn_info;
+
+	if (ioctl(dev, HCIGETCONNLIST, (void *) cl) < 0) {
+		g_free(cl);
+		return -EIO;
+	}
+
+	for (i = 0; i < cl->conn_num; i++, ci++) {
+		if (bacmp(&ci->bdaddr, dba) == 0) {
+			*handle = ci->handle;
+			g_free(cl);
+			return 0;
+		}
+	}
+
+	g_free(cl);
+
+	return -ENOENT;
+}
+
 static inline int get_bdaddr(int dev, bdaddr_t *sba, uint16_t handle, bdaddr_t *dba)
 {
 	struct hci_conn_list_req *cl;
@@ -244,14 +276,28 @@ static inline void update_lastused(bdaddr_t *sba, bdaddr_t *dba)
 
 static void link_key_request(int dev, bdaddr_t *sba, bdaddr_t *dba)
 {
+	struct hci_auth_info_req req;
 	unsigned char key[16];
 	char sa[18], da[18];
+	uint8_t type;
 	int err;
 
 	ba2str(sba, sa); ba2str(dba, da);
 	info("link_key_request (sba=%s, dba=%s)", sa, da);
 
-	err = read_link_key(sba, dba, key);
+	memset(&req, 0, sizeof(req));
+	bacpy(&req.bdaddr, dba);
+
+	err = ioctl(dev, HCIGETAUTHINFO, (unsigned long) &req);
+	if (err < 0 && errno != EINVAL)
+		debug("HCIGETAUTHINFO failed %s (%d)",
+					strerror(errno), errno);
+	else
+		req.type = 0x00;
+
+	debug("kernel auth requirements = 0x%02x", req.type);
+
+	err = read_link_key(sba, dba, key, &type);
 	if (err < 0) {
 		/* Link key not found */
 		hci_send_cmd(dev, OGF_LINK_CTL, OCF_LINK_KEY_NEG_REPLY, 6, dba);
@@ -260,7 +306,14 @@ static void link_key_request(int dev, bdaddr_t *sba, bdaddr_t *dba)
 		link_key_reply_cp lr;
 		memcpy(lr.link_key, key, 16);
 		bacpy(&lr.bdaddr, dba);
-		hci_send_cmd(dev, OGF_LINK_CTL, OCF_LINK_KEY_REPLY,
+
+		debug("stored link key type = 0x%02x", type);
+
+		if ((type == 0x03 || type == 0x04) && (req.type & 0x01))
+			hci_send_cmd(dev, OGF_LINK_CTL,
+					OCF_LINK_KEY_NEG_REPLY, 6, dba);
+		else
+			hci_send_cmd(dev, OGF_LINK_CTL, OCF_LINK_KEY_REPLY,
 						LINK_KEY_REPLY_CP_SIZE, &lr);
 	}
 }
@@ -270,17 +323,34 @@ static void link_key_notify(int dev, bdaddr_t *sba, void *ptr)
 	evt_link_key_notify *evt = ptr;
 	bdaddr_t *dba = &evt->bdaddr;
 	char sa[18], da[18];
-	int dev_id;
+	int dev_id, err;
 
 	ba2str(sba, sa); ba2str(dba, da);
 	info("link_key_notify (sba=%s, dba=%s)", sa, da);
 
 	dev_id = hci_devid(sa);
 
-	write_link_key(sba, dba, evt->link_key, evt->key_type,
+	err = write_link_key(sba, dba, evt->link_key, evt->key_type,
 						io_data[dev_id].pin_length);
+	if (err < 0) {
+		uint16_t handle;
 
-	hcid_dbus_bonding_process_complete(sba, dba, 0);
+		error("write_link_key: %s (%d)", strerror(-err), -err);
+
+		hcid_dbus_bonding_process_complete(sba, dba, HCI_MEMORY_FULL);
+
+		if (get_handle(dev, sba, dba, &handle) == 0) {
+			disconnect_cp cp;
+
+			memset(&cp, 0, sizeof(cp));
+			cp.handle = htobs(handle);
+			cp.reason = HCI_OE_LOW_RESOURCES;
+
+			hci_send_cmd(dev, OGF_LINK_CTL, OCF_DISCONNECT,
+						DISCONNECT_CP_SIZE, &cp);
+		}
+	} else
+		hcid_dbus_bonding_process_complete(sba, dba, 0);
 
 	io_data[dev_id].pin_length = -1;
 }
@@ -311,12 +381,28 @@ static void return_link_keys(int dev, bdaddr_t *sba, void *ptr)
 
 static void user_confirm_request(int dev, bdaddr_t *sba, void *ptr)
 {
-	hci_send_cmd(dev, OGF_LINK_CTL, OCF_USER_CONFIRM_NEG_REPLY, 6, ptr);
+	evt_user_confirm_request *req = ptr;
+
+	if (hcid_dbus_user_confirm(sba, &req->bdaddr,
+					btohl(req->passkey)) < 0)
+		hci_send_cmd(dev, OGF_LINK_CTL,
+				OCF_USER_CONFIRM_NEG_REPLY, 6, ptr);
 }
 
 static void user_passkey_request(int dev, bdaddr_t *sba, void *ptr)
 {
-	hci_send_cmd(dev, OGF_LINK_CTL, OCF_USER_PASSKEY_NEG_REPLY, 6, ptr);
+	evt_user_passkey_request *req = ptr;
+
+	if (hcid_dbus_user_passkey(sba, &req->bdaddr) < 0)
+		hci_send_cmd(dev, OGF_LINK_CTL,
+				OCF_USER_PASSKEY_NEG_REPLY, 6, ptr);
+}
+
+static void user_passkey_notify(int dev, bdaddr_t *sba, void *ptr)
+{
+	evt_user_passkey_notify *req = ptr;
+
+	hcid_dbus_user_notify(sba, &req->bdaddr, btohl(req->passkey));
 }
 
 static void remote_oob_data_request(int dev, bdaddr_t *sba, void *ptr)
@@ -326,18 +412,41 @@ static void remote_oob_data_request(int dev, bdaddr_t *sba, void *ptr)
 
 static void io_capa_request(int dev, bdaddr_t *sba, bdaddr_t *dba)
 {
-	io_capability_neg_reply_cp cp;
 	char sa[18], da[18];
+	uint8_t cap, auth;
 
 	ba2str(sba, sa); ba2str(dba, da);
 	info("io_capa_request (sba=%s, dba=%s)", sa, da);
 
-	memset(&cp, 0, sizeof(cp));
-	bacpy(&cp.bdaddr, dba);
-	cp.reason = HCI_PAIRING_NOT_ALLOWED;
-
-	hci_send_cmd(dev, OGF_LINK_CTL, OCF_IO_CAPABILITY_NEG_REPLY,
+	if (hcid_dbus_get_io_cap(sba, dba, &cap, &auth) < 0) {
+		io_capability_neg_reply_cp cp;
+		memset(&cp, 0, sizeof(cp));
+		bacpy(&cp.bdaddr, dba);
+		cp.reason = HCI_PAIRING_NOT_ALLOWED;
+		hci_send_cmd(dev, OGF_LINK_CTL, OCF_IO_CAPABILITY_NEG_REPLY,
 					IO_CAPABILITY_NEG_REPLY_CP_SIZE, &cp);
+	} else {
+		io_capability_reply_cp cp;
+		memset(&cp, 0, sizeof(cp));
+		bacpy(&cp.bdaddr, dba);
+		cp.capability = cap;
+		cp.oob_data = 0x00;
+		cp.authentication = auth;
+		hci_send_cmd(dev, OGF_LINK_CTL, OCF_IO_CAPABILITY_REPLY,
+					IO_CAPABILITY_REPLY_CP_SIZE, &cp);
+	}
+}
+
+static void io_capa_response(int dev, bdaddr_t *sba, void *ptr)
+{
+	evt_io_capability_response *evt = ptr;
+	char sa[18], da[18];
+
+	ba2str(sba, sa); ba2str(&evt->bdaddr, da);
+	info("io_capa_response (sba=%s, dba=%s)", sa, da);
+
+	hcid_dbus_set_io_cap(sba, &evt->bdaddr,
+				evt->capability, evt->authentication);
 }
 
 /* PIN code handling */
@@ -382,7 +491,7 @@ static void pin_code_request(int dev, bdaddr_t *sba, bdaddr_t *dba)
 	pinlen = read_pin_code(sba, dba, pin);
 
 	if (pairing == HCID_PAIRING_ONCE) {
-		err = read_link_key(sba, dba, key);
+		err = read_link_key(sba, dba, key, NULL);
 		if (!err) {
 			ba2str(dba, da);
 			error("PIN code request for already paired device %s", da);
@@ -391,39 +500,26 @@ static void pin_code_request(int dev, bdaddr_t *sba, bdaddr_t *dba)
 	} else if (pairing == HCID_PAIRING_NONE)
 		goto reject;
 
-	if (hcid.security == HCID_SEC_AUTO) {
-		if (!ci->out) {
-			/* Incomming connection */
-			set_pin_length(sba, hcid.pin_len);
-			memcpy(pr.pin_code, hcid.pin_code, hcid.pin_len);
-			pr.pin_len = hcid.pin_len;
-			hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
-				PIN_CODE_REPLY_CP_SIZE, &pr);
-		} else {
-			/* Outgoing connection */
-			if (pinlen > 0) {
-				set_pin_length(sba, pinlen);
-				memcpy(pr.pin_code, pin, pinlen);
-				pr.pin_len = pinlen;
-				hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
-					PIN_CODE_REPLY_CP_SIZE, &pr);
-			} else {
-				/* Request PIN from passkey agent */
-				hcid_dbus_request_pin(dev, sba, ci);
-			}
-		}
+	if (hcid.security == HCID_SEC_AUTO && !ci->out) {
+		set_pin_length(sba, hcid.pin_len);
+		memcpy(pr.pin_code, hcid.pin_code, hcid.pin_len);
+		pr.pin_len = hcid.pin_len;
+		hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
+						PIN_CODE_REPLY_CP_SIZE, &pr);
 	} else {
 		if (pinlen > 0) {
-			/* Confirm PIN by passkey agent */
-			hcid_dbus_confirm_pin(dev, sba, ci, pin);
+			set_pin_length(sba, pinlen);
+			memcpy(pr.pin_code, pin, pinlen);
+			pr.pin_len = pinlen;
+			hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_REPLY,
+						PIN_CODE_REPLY_CP_SIZE, &pr);
 		} else {
-			/* Request PIN from passkey agent */ 
-			hcid_dbus_request_pin(dev, sba, ci);
+			/* Request PIN from passkey agent */
+			if (hcid_dbus_request_pin(dev, sba, ci) < 0)
+				goto reject;
 		}
 	}
 
-	hcid_dbus_pending_pin_req_add(sba, &ci->bdaddr);
-
 	g_free(cr);
 
 	return;
@@ -432,8 +528,6 @@ reject:
 	g_free(cr);
 
 	hci_send_cmd(dev, OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY, 6, dba);
-
-	return;
 }
 
 static inline void cmd_status(int dev, bdaddr_t *sba, void *ptr)
@@ -478,10 +572,6 @@ static inline void cmd_complete(int dev, bdaddr_t *sba, void *ptr)
 	case cmd_opcode_pack(OGF_HOST_CTL, OCF_WRITE_SIMPLE_PAIRING_MODE):
 		hcid_dbus_write_simple_pairing_mode_complete(sba);
 		break;
-	case cmd_opcode_pack(OGF_LINK_CTL, OCF_PIN_CODE_REPLY):
-	case cmd_opcode_pack(OGF_LINK_CTL, OCF_PIN_CODE_NEG_REPLY):
-		hcid_dbus_pin_code_reply(sba, ptr);
-		break;
 	};
 }
 
@@ -630,7 +720,8 @@ static inline void conn_complete(int dev, int dev_id, bdaddr_t *sba, void *ptr)
 	if (evt->link_type != ACL_LINK)
 		return;
 
-	hcid_dbus_conn_complete(sba, evt->status, evt->handle, &evt->bdaddr);
+	hcid_dbus_conn_complete(sba, evt->status, btohs(evt->handle),
+				&evt->bdaddr);
 
 	if (evt->status)
 		return;
@@ -658,6 +749,7 @@ static inline void conn_complete(int dev, int dev_id, bdaddr_t *sba, void *ptr)
 	if (!str) {
 		read_remote_version_cp cp;
 
+		memset(&cp, 0, sizeof(cp));
 		cp.handle = evt->handle;
 
 		data = hci_req_data_new(dev_id, &evt->bdaddr, OGF_LINK_CTL,
@@ -673,7 +765,8 @@ static inline void disconn_complete(int dev, bdaddr_t *sba, void *ptr)
 {
 	evt_disconn_complete *evt = ptr;
 
-	hcid_dbus_disconn_complete(sba, evt->status, evt->handle, evt->reason);
+	hcid_dbus_disconn_complete(sba, evt->status, btohs(evt->handle),
+					evt->reason);
 }
 
 static inline void auth_complete(int dev, bdaddr_t *sba, void *ptr)
@@ -681,7 +774,7 @@ static inline void auth_complete(int dev, bdaddr_t *sba, void *ptr)
 	evt_auth_complete *evt = ptr;
 	bdaddr_t dba;
 
-	if (get_bdaddr(dev, sba, btohs(evt->handle), &dba) < 0) 
+	if (get_bdaddr(dev, sba, btohs(evt->handle), &dba) < 0)
 		return;
 
 	if (evt->status)
@@ -816,6 +909,10 @@ static gboolean io_security_event(GIOChannel *chan, GIOCondition cond, gpointer
 		io_capa_request(dev, &di->bdaddr, (bdaddr_t *) ptr);
 		break;
 
+	case EVT_IO_CAPABILITY_RESPONSE:
+		io_capa_response(dev, &di->bdaddr, ptr);
+		break;
+
 	case EVT_USER_CONFIRM_REQUEST:
 		user_confirm_request(dev, &di->bdaddr, ptr);
 		break;
@@ -824,6 +921,10 @@ static gboolean io_security_event(GIOChannel *chan, GIOCondition cond, gpointer
 		user_passkey_request(dev, &di->bdaddr, ptr);
 		break;
 
+	case EVT_USER_PASSKEY_NOTIFY:
+		user_passkey_notify(dev, &di->bdaddr, ptr);
+		break;
+
 	case EVT_REMOTE_OOB_DATA_REQUEST:
 		remote_oob_data_request(dev, &di->bdaddr, ptr);
 		break;
