@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -39,11 +39,9 @@
 #include <bluetooth/sdp_lib.h>
 
 #include <glib.h>
-
 #include <dbus/dbus.h>
+#include <gdbus.h>
 
-#include "dbus.h"
-#include "dbus-helper.h"
 #include "logging.h"
 #include "textfile.h"
 
@@ -52,10 +50,7 @@
 #include "error.h"
 #include "manager.h"
 #include "storage.h"
-
-static const char *pnp_uuid	= "00001200-0000-1000-8000-00805f9b34fb";
-static const char *hid_uuid	= "00001124-0000-1000-8000-00805f9b34fb";
-static const char *headset_uuid	= "00001108-0000-1000-8000-00805f9b34fb";
+#include "glib-helper.h"
 
 struct pending_req {
 	char		*adapter_path;	/* Local adapter D-Bus path */
@@ -63,11 +58,13 @@ struct pending_req {
 	bdaddr_t	dst;		/* Peer BT address */
 	DBusConnection	*conn;
 	DBusMessage	*msg;
-	sdp_record_t	*pnp_rec;
-	sdp_record_t	*hid_rec;
-	int		ctrl_sock;
+	sdp_list_t	*pnp_recs;
+	sdp_list_t	*hid_recs;
+	GIOChannel	*ctrl_channel;
 };
 
+static int idle_timeout = 0;
+
 static GSList *device_paths = NULL;	/* Input registered paths */
 
 static DBusConnection *connection = NULL;
@@ -110,79 +107,15 @@ static void pending_req_free(struct pending_req *pr)
 	if (pr->msg)
 		dbus_message_unref(pr->msg);
 
-	if (pr->pnp_rec)
-		sdp_record_free(pr->pnp_rec);
+	if (pr->pnp_recs)
+		sdp_list_free(pr->pnp_recs, (sdp_free_func_t) sdp_record_free);
 
-	if (pr->hid_rec)
-		sdp_record_free(pr->hid_rec);
+	if (pr->hid_recs)
+		sdp_list_free(pr->hid_recs, (sdp_free_func_t) sdp_record_free);
 
 	g_free(pr);
 }
 
-static int get_record(struct pending_req *pr, uint32_t handle,
-					DBusPendingCallNotifyFunction cb)
-{
-	DBusMessage *msg;
-	DBusPendingCall *pending;
-	char addr[18];
-	const char *paddr = addr;
-
-	msg = dbus_message_new_method_call("org.bluez", pr->adapter_path,
-			"org.bluez.Adapter", "GetRemoteServiceRecord");
-	if (!msg)
-		return -1;
-
-	ba2str(&pr->dst, addr);
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &paddr,
-			DBUS_TYPE_UINT32, &handle,
-			DBUS_TYPE_INVALID);
-
-	if (dbus_connection_send_with_reply(pr->conn, msg, &pending, -1) == FALSE) {
-		error("Can't send D-Bus message.");
-		dbus_message_unref(msg);
-		return -1;
-	}
-
-	dbus_pending_call_set_notify(pending, cb, pr, NULL);
-	dbus_pending_call_unref(pending);
-	dbus_message_unref(msg);
-
-	return 0;
-}
-
-static int get_handles(struct pending_req *pr, const char *uuid,
-					DBusPendingCallNotifyFunction cb)
-{
-	DBusMessage *msg;
-	DBusPendingCall *pending;
-	char addr[18];
-	const char *paddr = addr;
-
-	msg = dbus_message_new_method_call("org.bluez", pr->adapter_path,
-			"org.bluez.Adapter", "GetRemoteServiceHandles");
-	if (!msg)
-		return -1;
-
-	ba2str(&pr->dst, addr);
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &paddr,
-			DBUS_TYPE_STRING, &uuid,
-			DBUS_TYPE_INVALID);
-	
-	if (dbus_connection_send_with_reply(pr->conn, msg, &pending, -1) == FALSE) {
-		error("Can't send D-Bus message.");
-		dbus_message_unref(msg);
-		return -1;
-	}
-
-	dbus_pending_call_set_notify(pending, cb, pr, NULL);
-	dbus_pending_call_unref(pending);
-	dbus_message_unref(msg);
-
-	return 0;
-}
-
 static void epox_endian_quirk(unsigned char *data, int size)
 {
 	/* USAGE_PAGE (Keyboard)	05 07
@@ -279,145 +212,88 @@ static void extract_pnp_record(sdp_record_t *rec, struct hidp_connadd_req *req)
 	req->version = pdlist ? pdlist->val.uint16 : 0x0000;
 }
 
-static gboolean interrupt_connect_cb(GIOChannel *chan,
-			GIOCondition cond, struct pending_req *pr)
+static void interrupt_connect_cb(GIOChannel *chan, int err,
+			const bdaddr_t *src, const bdaddr_t *dst,
+			gpointer user_data)
 {
+	struct pending_req *pr = user_data;
 	struct hidp_connadd_req hidp;
-	DBusMessage *reply;
 	const char *path;
-	int isk, ret, err;
-	socklen_t len;
 
-	isk = g_io_channel_unix_get_fd(chan);
+	memset(&hidp, 0, sizeof(hidp));
 
-	if (cond & G_IO_NVAL) {
-		err = EHOSTDOWN;
-		isk = -1;
+	if (err < 0) {
+		error("connect(): %s (%d)", strerror(-err), -err);
 		goto failed;
 	}
 
-	if (cond & (G_IO_HUP | G_IO_ERR)) {
-		err = EHOSTDOWN;
-		error("Hangup or error on HIDP interrupt socket");
-		goto failed;
+	g_io_channel_close(chan);
+	g_io_channel_unref(chan);
 
-	}
+	hidp.idle_to = idle_timeout * 60;
 
-	len = sizeof(ret);
-	if (getsockopt(isk, SOL_SOCKET, SO_ERROR, &ret, &len) < 0) {
-		err = errno;
-		error("getsockopt(SO_ERROR): %s (%d)", strerror(err), err);
-		goto failed;
-	}
-
-	if (ret != 0) {
-		err = ret;
-		error("connect(): %s (%d)", strerror(ret), ret);
-		goto failed;
-	}
-
-	memset(&hidp, 0, sizeof(struct hidp_connadd_req));
-	extract_hid_record(pr->hid_rec, &hidp);
-	if (pr->pnp_rec)
-		extract_pnp_record(pr->pnp_rec, &hidp);
+	extract_hid_record(pr->hid_recs->data, &hidp);
+	if (pr->pnp_recs)
+		extract_pnp_record(pr->pnp_recs->data, &hidp);
 
 	store_device_info(&pr->src, &pr->dst, &hidp);
 
 	if (input_device_register(pr->conn, &pr->src,
 					&pr->dst, &hidp, &path) < 0) {
-		err_failed(pr->conn, pr->msg, "path registration failed");
+		error_failed(pr->conn, pr->msg, "path registration failed");
 		goto cleanup;
 	}
 
-	dbus_connection_emit_signal(pr->conn, INPUT_PATH,
+	g_dbus_emit_signal(pr->conn, INPUT_PATH,
 			INPUT_MANAGER_INTERFACE, "DeviceCreated",
 			DBUS_TYPE_STRING, &path,
 			DBUS_TYPE_INVALID);
 
 	device_paths = g_slist_append(device_paths, g_strdup(path));
 
-	/* Replying to the requestor */
-	reply = dbus_message_new_method_return(pr->msg);
-
-	dbus_message_append_args(reply,
-				DBUS_TYPE_STRING, &path,
-				DBUS_TYPE_INVALID);
-
-	send_message_and_unref(pr->conn, reply);
+	g_dbus_send_reply(pr->conn, pr->msg,  DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 
 	goto cleanup;
+
 failed:
-	err_connection_failed(pr->conn, pr->msg, strerror(err));
+	error_connection_attempt_failed(pr->conn, pr->msg, err);
 
 cleanup:
-	if (isk >= 0)
-		close(isk);
-
-	close(pr->ctrl_sock);
+	g_io_channel_close(pr->ctrl_channel);
+	g_io_channel_unref(pr->ctrl_channel);
 	pending_req_free(pr);
 
 	if (hidp.rd_data)
 		g_free(hidp.rd_data);
-
-	return FALSE;
 }
 
-static gboolean control_connect_cb(GIOChannel *chan,
-			GIOCondition cond, struct pending_req *pr)
+static void control_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
+			const bdaddr_t *dst, gpointer user_data)
 {
-	int ret, csk, err;
-	socklen_t len;
+	struct pending_req *pr = user_data;
 
-	csk = g_io_channel_unix_get_fd(chan);
-
-	if (cond & G_IO_NVAL) {
-		err = EHOSTDOWN;
-		csk = -1;
+	if (err < 0) {
+		error("connect(): %s (%d)", strerror(-err), -err);
 		goto failed;
 	}
 
-	if (cond & (G_IO_HUP | G_IO_ERR)) {
-		err = EHOSTDOWN;
-		error("Hangup or error on HIDP control socket");
-		goto failed;
-
-	}
-
 	/* Set HID control channel */
-	pr->ctrl_sock = csk;
-
-	len = sizeof(ret);
-	if (getsockopt(csk, SOL_SOCKET, SO_ERROR, &ret, &len) < 0) {
-		err = errno;
-		error("getsockopt(SO_ERROR): %s (%d)", strerror(err), err);
-		goto failed;
-	}
-
-	if (ret != 0) {
-		err = ret;
-		error("connect(): %s (%d)", strerror(ret), ret);
-		goto failed;
-	}
+	pr->ctrl_channel = chan;
 
 	/* Connect to the HID interrupt channel */
-	if (l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_INTR,
-			(GIOFunc) interrupt_connect_cb, pr) < 0) {
-
-		err = errno;
-		error("L2CAP connect failed:%s (%d)", strerror(errno), errno);
+	err = bt_l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_INTR, 0,
+			interrupt_connect_cb, pr);
+	if (err < 0) {
+		error("L2CAP connect failed:%s (%d)", strerror(-err), -err);
 		goto failed;
 	}
 
-	return FALSE;
+	return;
 
 failed:
-	if (csk >= 0)
-		close(csk);
-
-	err_connection_failed(pr->conn, pr->msg, strerror(err));
+	error_connection_attempt_failed(pr->conn, pr->msg, -err);
 	pending_req_free(pr);
-
-	return FALSE;
 }
 
 static void create_bonding_reply(DBusPendingCall *call, void *data)
@@ -425,12 +301,13 @@ static void create_bonding_reply(DBusPendingCall *call, void *data)
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	struct pending_req *pr = data;
 	DBusError derr;
+	int err;
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
 		error("CreateBonding failed: %s(%s)",
 					derr.name, derr.message);
-		err_authentication_failed(pr->conn, pr->msg);
+		error_failed(pr->conn, pr->msg, "Authentication failed (CreateBonding)");
 		dbus_error_free(&derr);
 		dbus_message_unref(reply);
 		pending_req_free(pr);
@@ -439,50 +316,15 @@ static void create_bonding_reply(DBusPendingCall *call, void *data)
 
 	dbus_message_unref(reply);
 
-	if (l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_CTRL,
-				(GIOFunc) control_connect_cb, pr) < 0) {
-		int err = errno;
-		err_connection_failed(pr->conn, pr->msg, strerror(err));
-		error("L2CAP connect failed:%s (%d)", strerror(err), err);
+	err = bt_l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_CTRL, 0,
+			control_connect_cb, pr);
+	if (err < 0) {
+		error("L2CAP connect failed:%s (%d)", strerror(-err), -err);
+		error_connection_attempt_failed(pr->conn, pr->msg, -err);
 		pending_req_free(pr);
 	}
 }
 
-static void finish_sdp_transaction(bdaddr_t *dba)
-{
-	char address[18], *addr_ptr = address;
-	DBusMessage *msg, *reply;
-	DBusError derr;
-
-	ba2str(dba, address);
-
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez/hci0",
-						"org.bluez.Adapter",
-						"FinishRemoteServiceTransaction");
-	if (!msg) {
-		error("Unable to allocate new method call");
-		return;
-	}
-
-	dbus_message_append_args(msg, DBUS_TYPE_STRING, &addr_ptr,
-					DBUS_TYPE_INVALID);
-
-	dbus_error_init(&derr);
-	reply = dbus_connection_send_with_reply_and_block(connection, msg,
-								-1, &derr);
-
-	dbus_message_unref(msg);
-
-	if (dbus_error_is_set(&derr) || dbus_set_error_from_message(&derr, reply)) {
-		error("FinishRemoteServiceTransaction(%s) failed: %s",
-				address, derr.message);
-		dbus_error_free(&derr);
-		return;
-	}
-
-	dbus_message_unref(reply);
-}
-
 static int create_bonding(struct pending_req *pr)
 {
 	DBusPendingCall *pending;
@@ -490,7 +332,7 @@ static int create_bonding(struct pending_req *pr)
 	char address[18], *addr_ptr = address;
 
 	msg = dbus_message_new_method_call("org.bluez", pr->adapter_path,
-					   "org.bluez.Adapter", "CreateBonding");
+					"org.bluez.Adapter", "CreateBonding");
 	if (!msg) {
 		error("Unable to allocate new method call");
 		return -1;
@@ -509,60 +351,35 @@ static int create_bonding(struct pending_req *pr)
 	return 0;
 }
 
-static void hid_record_reply(DBusPendingCall *call, void *data)
+static void hid_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct pending_req *pr = data;
-	DBusError derr;
-	uint8_t *rec_bin;
-	int len, scanned;
+	struct pending_req *pr = user_data;
 
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceRecord failed: %s(%s)",
-					derr.name, derr.message);
+	if (err < 0) {
+		error_not_supported(pr->conn, pr->msg);
+		error("SDP search error: %s (%d)", strerror(-err), -err);
 		goto fail;
 	}
 
-	finish_sdp_transaction(&pr->dst);
-
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &rec_bin, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
+	if (!recs || !recs->data) {
+		error_not_supported(pr->conn, pr->msg);
 		error("Invalid HID service record length");
 		goto fail;
 	}
 
-	pr->hid_rec = sdp_extract_pdu(rec_bin, &scanned);
-	if (!pr->hid_rec) {
-		err_not_supported(pr->conn, pr->msg);
-		goto fail;
-	}
-
-	dbus_message_unref(reply);
+	pr->hid_recs = recs;
 
 	if (strcmp("CreateSecureDevice", dbus_message_get_member(pr->msg)) == 0) {
 		sdp_data_t *d;
 
 		/* Pairing mandatory for keyboard and combo */
-		d = sdp_data_get(pr->hid_rec, SDP_ATTR_HID_DEVICE_SUBCLASS);
+		d = sdp_data_get(pr->hid_recs->data,
+				SDP_ATTR_HID_DEVICE_SUBCLASS);
 		if (d && (d->val.uint8 & 0x40) &&
 				!has_bonding(&pr->src, &pr->dst)) {
 			if (create_bonding(pr) < 0) {
-				err_authentication_failed(pr->conn, pr->msg);
+				error_failed(pr->conn, pr->msg,
+					"Unable to initialize bonding process");
 				goto fail;
 			}
 			/* Wait bonding reply */
@@ -573,231 +390,78 @@ static void hid_record_reply(DBusPendingCall *call, void *data)
 	}
 
 	/* No encryption or link key already exists -- connect control channel */
-	if (l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_CTRL,
-				(GIOFunc) control_connect_cb, pr) < 0) {
-		int err = errno;
-		error("L2CAP connect failed:%s (%d)", strerror(err), err);
-		err_connection_failed(pr->conn, pr->msg, strerror(err));
+	err = bt_l2cap_connect(&pr->src, &pr->dst, L2CAP_PSM_HIDP_CTRL, 0,
+			control_connect_cb, pr);
+	if (err < 0) {
+		error("L2CAP connect failed:%s (%d)", strerror(-err), -err);
+		error_connection_attempt_failed(pr->conn, pr->msg, -err);
 		goto fail;
 	}
 
 	/* Wait L2CAP connect */
 	return;
-fail:
-	dbus_error_free(&derr);
-	pending_req_free(pr);
-	dbus_message_unref(reply);
-}
-
-static void hid_handle_reply(DBusPendingCall *call, void *data)
-{
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct pending_req *pr = data;
-	uint32_t *phandle;
-	DBusError derr;
-	int len;
-
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceHandles: %s(%s)",
-					derr.name, derr.message);
-		goto fail;
-	}
-
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &phandle, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
-		error("HID record handle not found");
-		goto fail;
-	}
-
-	if (get_record(pr, *phandle, hid_record_reply) < 0) {
-		err_not_supported(pr->conn, pr->msg);
-		error("HID service attribute request failed");
-		goto fail;
-	} else {
-		/* Wait record reply */
-		goto done;
-	}
 
 fail:
-	dbus_error_free(&derr);
 	pending_req_free(pr);
-
-done:
-	dbus_message_unref(reply);
 }
 
-static void pnp_record_reply(DBusPendingCall *call, void *data)
+static void pnp_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct pending_req *pr = data;
-	DBusError derr;
-	uint8_t *rec_bin;
-	int len, scanned;
+	struct pending_req *pr = user_data;
+	uuid_t uuid;
 
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceRecord: %s(%s)",
-				derr.name, derr.message);
+	if (err < 0) {
+		error_not_supported(pr->conn, pr->msg);
+		error("SDP search error: %s (%d)", strerror(-err), -err);
 		goto fail;
 	}
 
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &rec_bin, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
+	if (!recs || !recs->data) {
+		error_not_supported(pr->conn, pr->msg);
 		error("Invalid PnP service record length");
 		goto fail;
 	}
 
-	pr->pnp_rec = sdp_extract_pdu(rec_bin, &scanned);
-	if (get_handles(pr, hid_uuid, hid_handle_reply) < 0) {
-		err_not_supported(pr->conn, pr->msg);
+	pr->pnp_recs = recs;
+	sdp_uuid16_create(&uuid, HID_SVCLASS_ID);
+	err = bt_search_service(&pr->src, &pr->dst, &uuid, hid_record_cb,
+			pr, NULL);
+	if (err < 0) {
+		error_not_supported(pr->conn, pr->msg);
 		error("HID service search request failed");
 		goto fail;
-	} else {
-		/* Wait handle reply */
-		goto done;
 	}
 
+	return;
+
 fail:
-	dbus_error_free(&derr);
 	pending_req_free(pr);
-
-done:
-	dbus_message_unref(reply);
 }
 
-static void pnp_handle_reply(DBusPendingCall *call, void *data)
+static void headset_record_cb(sdp_list_t *recs, int err, gpointer user_data)
 {
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct pending_req *pr = data;
-	DBusError derr;
-	uint32_t *phandle;
-	int len;
-
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceHandles: %s(%s)",
-				derr.name, derr.message);
-		goto fail;
-	}
-
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &phandle, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		/* PnP is optional: Ignore it and request the HID handle  */
-		if (get_handles(pr, hid_uuid, hid_handle_reply) < 0) {
-			err_not_supported(pr->conn, pr->msg);
-			error("HID service search request failed");
-			goto fail;
-		}
-	} else {
-		/* Request PnP record */
-		if (get_record(pr, *phandle, pnp_record_reply) < 0) {
-			err_not_supported(pr->conn, pr->msg);
-			error("PnP service attribute request failed");
-			goto fail;
-		}
-	}
-
-	/* Wait HID handle reply or PnP record reply */
-	goto done;
-
-fail:
-	dbus_error_free(&derr);
-	pending_req_free(pr);
-
-done:
-	dbus_message_unref(reply);
-}
-
-static void headset_record_reply(DBusPendingCall *call, void *data)
-{
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	DBusMessage *pr_reply;
-	DBusError derr;
-	struct pending_req *pr = data;
-	uint8_t *rec_bin;
+	struct pending_req *pr = user_data;
 	sdp_record_t *rec;
 	sdp_list_t *protos;
 	const char *path;
-	int len, scanned;
 	uint8_t ch;
 
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceRecord: %s(%s)",
-				derr.name, derr.message);
+	if (err < 0) {
+		error_not_supported(pr->conn, pr->msg);
+		error("SDP search error: %s (%d)", strerror(-err), -err);
 		goto fail;
 	}
 
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &rec_bin, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
+	if (!recs || !recs->data) {
+		error_not_supported(pr->conn, pr->msg);
 		error("Invalid headset service record length");
 		goto fail;
 	}
 
-	rec = sdp_extract_pdu(rec_bin, &scanned);
-	if (!rec) {
-		err_not_supported(pr->conn, pr->msg);
-		goto fail;
-	}
+	rec = recs->data;
 
 	if (sdp_get_access_protos(rec, &protos) < 0) {
-		err_not_supported(pr->conn, pr->msg);
+		error_not_supported(pr->conn, pr->msg);
 		goto fail;
 	}
 
@@ -807,7 +471,7 @@ static void headset_record_reply(DBusPendingCall *call, void *data)
 	sdp_record_free(rec);
 
 	if (ch <= 0) {
-		err_not_supported(pr->conn, pr->msg);
+		error_not_supported(pr->conn, pr->msg);
 		error("Invalid RFCOMM channel");
 		goto fail;
 	}
@@ -816,189 +480,138 @@ static void headset_record_reply(DBusPendingCall *call, void *data)
 
 	if (fake_input_register(pr->conn, &pr->src, &pr->dst, ch, &path) < 0) {
 		error("D-Bus path registration failed:%s", path);
-		err_failed(pr->conn, pr->msg, "Path registration failed");
+		error_failed(pr->conn, pr->msg, "Path registration failed");
 		goto fail;
 	}
 
-	dbus_connection_emit_signal(pr->conn, INPUT_PATH,
+	g_dbus_emit_signal(pr->conn, INPUT_PATH,
 			INPUT_MANAGER_INTERFACE, "DeviceCreated",
 			DBUS_TYPE_STRING, &path,
 			DBUS_TYPE_INVALID);
 
 	device_paths = g_slist_append(device_paths, g_strdup(path));
 
-	pr_reply = dbus_message_new_method_return(pr->msg);
-
-	dbus_message_append_args(pr_reply,
-			DBUS_TYPE_STRING, &path,
-			DBUS_TYPE_INVALID);
-
-	send_message_and_unref(pr->conn, pr_reply);
+	g_dbus_send_reply(pr->conn, pr->msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 
 fail:
-	dbus_error_free(&derr);
 	pending_req_free(pr);
-	dbus_message_unref(reply);
 }
 
-static void headset_handle_reply(DBusPendingCall *call, void *data)
+static inline DBusMessage *adapter_not_available(DBusMessage *msg)
 {
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	struct pending_req *pr = data;
-	DBusError derr;
-	uint32_t *phandle;
-	int len;
-
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-			"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceHandles: %s(%s)",
-				derr.name, derr.message);
-		goto fail;
-	}
-
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &phandle, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
-		error("Headset record handle not found");
-		goto fail;
-	}
-
-	if (get_record(pr, *phandle, headset_record_reply) < 0) {
-		err_not_supported(pr->conn, pr->msg);
-		error("Headset service attribute request failed");
-		goto fail;
-	} else {
-		/* Wait record reply */
-		goto done;
-	}
-
-fail:
-	dbus_error_free(&derr);
-	pending_req_free(pr);
-
-done:
-	dbus_message_unref(reply);
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"Adapter not available");
 }
 
-static DBusHandlerResult create_device(DBusConnection *conn,
+static inline DBusMessage *already_exists(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExists",
+						"Input Already exists");
+}
+
+static inline DBusMessage *not_supported(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotSupported",
+							"Not supported");
+}
+
+static inline DBusMessage *does_not_exist(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExists",
+							"Input doesn't exist");
+}
+
+static DBusMessage *create_device(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct pending_req *pr;
-	DBusError derr;
 	const char *addr;
 	bdaddr_t src, dst;
 	uint32_t cls = 0;
-	int dev_id;
+	int dev_id, err;
+	uuid_t uuid;
+	bt_callback_t cb;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &addr,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &addr,
+							DBUS_TYPE_INVALID))
+		return NULL;
 
 	/* Get the default adapter */
 	dev_id = hci_get_route(NULL);
 	if (dev_id < 0) {
 		error("Bluetooth adapter not available");
-		return err_failed(conn, msg, "Adapter not available");
+		return adapter_not_available(msg);
 	}
 
 	if (hci_devba(dev_id, &src) < 0) {
 		error("Can't get local adapter device info");
-		return err_failed(conn, msg, "Adapter not available");
+		return adapter_not_available(msg);
 	}
 
 	str2ba(addr, &dst);
 	if (input_device_is_registered(&src, &dst))
-		return err_already_exists(conn, msg, "Input Already exists");
+		return already_exists(msg);
 
 	if (read_device_class(&src, &dst, &cls) < 0) {
 		error("Device class not available");
-		return err_not_supported(conn, msg);
+		return not_supported(msg);
 	}
 
 	pr = pending_req_new(conn, msg, &src, &dst);
 	if (!pr)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	switch (cls & 0x1f00) {
 		case 0x0500: /* Peripheral */
 		case 0x0200: /* Phone */
-			if (get_handles(pr, pnp_uuid, pnp_handle_reply) < 0) {
-				pending_req_free(pr);
-				return err_not_supported(conn, msg);
-			}
+			sdp_uuid16_create(&uuid, PNP_INFO_SVCLASS_ID);
+			cb = pnp_record_cb;
 			break;
 		case 0x0400: /* Fake input */
-			if (get_handles(pr, headset_uuid,
-						headset_handle_reply) < 0) {
-				pending_req_free(pr);
-				return err_not_supported(conn, msg);
-			}
+			sdp_uuid16_create(&uuid, HEADSET_SVCLASS_ID);
+			cb = headset_record_cb;
 			break;
 		default:
 			pending_req_free(pr);
-			return err_not_supported(conn, msg);
+			return not_supported(msg);
 	}
 
-	return DBUS_HANDLER_RESULT_HANDLED;
+	err = bt_search_service(&src, &dst, &uuid, cb, pr, NULL);
+	if (err < 0) {
+		pending_req_free(pr);
+		return not_supported(msg);
+	}
+
+	return NULL;
 }
 
-static DBusHandlerResult remove_device(DBusConnection *conn,
+static DBusMessage *remove_device(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	DBusMessage *reply;
-	DBusError derr;
 	GSList *l;
 	const char *path;
 	int err;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &path,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID))
+		return NULL;
 
 	l = g_slist_find_custom(device_paths, path, (GCompareFunc) strcmp);
 	if (!l)
-		return err_does_not_exist(conn, msg, "Input doesn't exist");
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return does_not_exist(msg);
 
 	err = input_device_unregister(conn, path);
-	if (err < 0) {
-		dbus_message_unref(reply);
-		return err_failed(conn, msg, strerror(-err));
-	}
+	if (err < 0)
+		return create_errno_message(msg, -err);
 
 	g_free(l->data);
 	device_paths = g_slist_remove(device_paths, l->data);
 
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult list_devices(DBusConnection *conn,
+static DBusMessage *list_devices(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	DBusMessageIter iter, iter_array;
@@ -1007,7 +620,7 @@ static DBusHandlerResult list_devices(DBusConnection *conn,
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	dbus_message_iter_init_append(reply, &iter);
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
@@ -1021,10 +634,10 @@ static DBusHandlerResult list_devices(DBusConnection *conn,
 
 	dbus_message_iter_close_container(&iter, &iter_array);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static void manager_unregister(DBusConnection *conn, void *data)
+static void manager_unregister(void *data)
 {
 	info("Unregistered manager path");
 
@@ -1045,7 +658,7 @@ static void stored_input(char *key, char *value, void *data)
 
 	str2ba(key, &dst);
 
-	memset(&hidp, 0, sizeof(struct hidp_connadd_req));
+	memset(&hidp, 0, sizeof(hidp));
 
 	if (parse_stored_device_info(value, &hidp) < 0)
 		return;
@@ -1080,7 +693,7 @@ static void stored_hidd(char *key, char *value, void *data)
 		return;
 	}
 
-	memset(&hidp, 0, sizeof(struct hidp_connadd_req));
+	memset(&hidp, 0, sizeof(hidp));
 
 	if (parse_stored_hidd(value, &hidp) < 0)
 		return;
@@ -1126,37 +739,40 @@ static void register_stored_inputs(void)
 	closedir(dir);
 }
 
-static DBusMethodVTable manager_methods[] = {
-	{ "ListDevices",		list_devices,	"",	"as"	},
-	{ "CreateDevice",		create_device,	"s",	"s"	},
-	{ "CreateSecureDevice",		create_device,	"s",	"s"	},
-	{ "RemoveDevice",		remove_device,	"s",	""	},
-	{ NULL, NULL, NULL, NULL },
+static GDBusMethodTable manager_methods[] = {
+	{ "ListDevices",	"",	"as",	list_devices },
+	{ "CreateDevice",	"s",	"s",	create_device,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "CreateSecureDevice",	"s",	"s",	create_device,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "RemoveDevice",	"s",	"",	remove_device },
+	{ }
 };
 
-static DBusSignalVTable manager_signals[] = {
+static GDBusSignalTable manager_signals[] = {
 	{ "DeviceCreated",	"s"	},
 	{ "DeviceRemoved",	"s"	},
-	{ NULL, NULL }
+	{ }
 };
 
-int input_init(DBusConnection *conn)
+int input_manager_init(DBusConnection *conn, GKeyFile *config)
 {
-	dbus_connection_set_exit_on_disconnect(conn, TRUE);
-
-	if (!dbus_connection_create_object_path(conn, INPUT_PATH,
-						NULL, manager_unregister)) {
-		error("D-Bus failed to register %s path", INPUT_PATH);
-		return -1;
+	GError *err = NULL;
+ 
+	if (config) {
+		idle_timeout = g_key_file_get_integer(config, "General",
+						"IdleTimeout", &err);
+		if (err) {
+			debug("input.conf: %s", err->message);
+			g_error_free(err);
+		}
 	}
 
-	if (!dbus_connection_register_interface(conn, INPUT_PATH,
-						INPUT_MANAGER_INTERFACE,
-						manager_methods,
-						manager_signals, NULL)) {
+	if (g_dbus_register_interface(conn, INPUT_PATH, INPUT_MANAGER_INTERFACE,
+					manager_methods, manager_signals, NULL,
+					NULL, manager_unregister) == FALSE) {
 		error("Failed to register %s interface to %s",
 				INPUT_MANAGER_INTERFACE, INPUT_PATH);
-		dbus_connection_destroy_object_path(connection, INPUT_PATH);
 		return -1;
 	}
 
@@ -1167,14 +783,15 @@ int input_init(DBusConnection *conn)
 	/* Register well known HID devices */
 	register_stored_inputs();
 
-	server_start(connection);
+	server_start();
 
 	return 0;
 }
 
-void input_exit(void)
+void input_manager_exit(void)
 {
-	dbus_connection_destroy_object_path(connection, INPUT_PATH);
+	g_dbus_unregister_interface(connection, INPUT_PATH,
+						INPUT_MANAGER_INTERFACE);
 
 	server_stop();
 
