@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -39,11 +39,11 @@
 #include <bluetooth/sdp_lib.h>
 
 #include <glib.h>
+#include <gdbus.h>
 
-#include "dbus.h"
-#include "dbus-helper.h"
 #include "logging.h"
 #include "textfile.h"
+#include "glib-helper.h"
 
 #define NETWORK_MANAGER_INTERFACE "org.bluez.network.Manager"
 
@@ -75,7 +75,6 @@ static DBusConnection *connection = NULL;
 
 static void pending_reply_free(struct pending_reply *pr)
 {
-
 	if (pr->addr)
 		g_free(pr->addr);
 	if (pr->path)
@@ -88,31 +87,52 @@ static void pending_reply_free(struct pending_reply *pr)
 		dbus_connection_unref(pr->conn);
 }
 
-static DBusHandlerResult create_path(DBusConnection *conn,
-					DBusMessage *msg, const char *path,
-					const char *sname)
+static inline DBusMessage *does_not_exist(DBusMessage *msg)
 {
-	DBusMessage *reply;
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExists",
+							"No such connection");
+}
 
+static inline DBusMessage *already_exists(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExists",
+						"Connection already exists");
+}
+
+static inline DBusMessage *not_supported(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotSupported",
+							"Not supported");
+}
+
+static inline DBusMessage *connection_is_busy(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"Connection is Busy");
+}
+
+static inline DBusMessage *adapter_not_available(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"Adapter not available");
+}
+
+static void create_path(DBusConnection *conn, DBusMessage *msg,
+				const char *path, const char *sname)
+{
 	/* emit signal when it is a new path */
 	if (sname) {
-		dbus_connection_emit_signal(conn, NETWORK_PATH,
+		g_dbus_emit_signal(conn, NETWORK_PATH,
 						NETWORK_MANAGER_INTERFACE,
 						sname, DBUS_TYPE_STRING, &path,
 						DBUS_TYPE_INVALID);
 	}
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	g_dbus_send_reply(conn, msg, DBUS_TYPE_STRING, &path,
+						DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult list_paths(DBusConnection *conn, DBusMessage *msg,
+static DBusMessage *list_paths(DBusConnection *conn, DBusMessage *msg,
 					GSList *list)
 {
 	DBusMessage *reply;
@@ -120,8 +140,8 @@ static DBusHandlerResult list_paths(DBusConnection *conn, DBusMessage *msg,
 	DBusMessageIter array_iter;
 
 	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+	if (reply == NULL)
+		return NULL;
 
 	dbus_message_iter_init_append(reply, &iter);
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
@@ -132,12 +152,13 @@ static DBusHandlerResult list_paths(DBusConnection *conn, DBusMessage *msg,
 						DBUS_TYPE_STRING,
 						&list->data);
 	}
+
 	dbus_message_iter_close_container(&iter, &array_iter);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static const char * last_connection_used(DBusConnection *conn)
+static const char *last_connection_used(DBusConnection *conn)
 {
 	const char *path = NULL;
 	GSList *l;
@@ -158,32 +179,25 @@ static const char * last_connection_used(DBusConnection *conn)
 	return path;
 }
 
-static DBusHandlerResult remove_path(DBusConnection *conn,
+static DBusMessage *remove_path(DBusConnection *conn,
 					DBusMessage *msg, GSList **list,
 					const char *sname)
 {
 	const char *path;
-	DBusMessage *reply;
-	DBusError derr;
 	GSList *l;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &path,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &path,
+						DBUS_TYPE_INVALID) == FALSE)
+		return NULL;
 
 	l = g_slist_find_custom(*list, path, (GCompareFunc) strcmp);
 	if (!l)
-		return err_does_not_exist(conn, msg, "Path doesn't exist");
+		return does_not_exist(msg);
 
 	/* Remove references from the storage */
 	if (*list == connection_paths) {
 		if (connection_has_pending(path))
-			return err_failed(conn, msg, "Connection is Busy");
+			return connection_is_busy(msg);
 
 		connection_remove_stored(path);
 		/* Reset default connection */
@@ -198,60 +212,36 @@ static DBusHandlerResult remove_path(DBusConnection *conn,
 	g_free(l->data);
 	*list = g_slist_remove(*list, l->data);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	if (!dbus_connection_destroy_object_path(conn, path))
-		error("Network path unregister failed");
-
-	dbus_connection_emit_signal(conn, NETWORK_PATH,
+	g_dbus_emit_signal(conn, NETWORK_PATH,
 					NETWORK_MANAGER_INTERFACE,
 					sname, DBUS_TYPE_STRING, &path,
 					DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	g_dbus_unregister_interface(conn, path, NETWORK_CONNECTION_INTERFACE);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static void pan_record_reply(DBusPendingCall *call, void *data)
+static void records_cb(sdp_list_t *recs, int err, gpointer data)
 {
 	struct pending_reply *pr = data;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	DBusError derr;
-	int len, scanned;
-	uint8_t *rec_bin;
+	int len;
 	sdp_data_t *d;
 	sdp_record_t *rec = NULL;
 	char name[MAX_NAME_SIZE], *desc = NULL;
 
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-				"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceRecord failed: %s(%s)", derr.name,
-			derr.message);
+	if (err < 0) {
+		error_connection_attempt_failed(pr->conn, pr->msg, -err);
 		goto fail;
 	}
 
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &rec_bin, &len,
-				DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
+	if (!recs || !recs->data) {
+		error_not_supported(pr->conn, pr->msg);
+		error("Invalid PAN service record");
 		goto fail;
 	}
 
-	if (len == 0) {
-		err_not_supported(pr->conn, pr->msg);
-		error("Invalid PAN service record length");
-		goto fail;
-	}
-
-	rec = sdp_extract_pdu(rec_bin, &scanned);
+	rec = recs->data;
 
 	/* Concat remote name and service name */
 	memset(name, 0, MAX_NAME_SIZE);
@@ -274,9 +264,11 @@ static void pan_record_reply(DBusPendingCall *call, void *data)
 				d->unitSize, d->val.str);
 	}
 
+	sdp_list_free(recs, (sdp_free_func_t) sdp_record_free);
+
 	if (connection_register(pr->path, &pr->src, &pr->dst, pr->id, name,
 				desc) < 0) {
-		err_failed(pr->conn, pr->msg, "D-Bus path registration failed");
+		error_failed(pr->conn, pr->msg, "D-Bus path registration failed");
 		goto fail;
 	}
 
@@ -284,146 +276,28 @@ static void pan_record_reply(DBusPendingCall *call, void *data)
 	connection_paths = g_slist_append(connection_paths, g_strdup(pr->path));
 
 	create_path(pr->conn, pr->msg, pr->path, "ConnectionCreated");
+
 fail:
-
-	if (desc)
-		g_free(desc);
-
-	sdp_record_free(rec);
-	dbus_error_free(&derr);
-	pending_reply_free(pr);
-	dbus_message_unref(reply);
-}
-
-static int get_record(struct pending_reply *pr, uint32_t handle,
-					DBusPendingCallNotifyFunction cb)
-{
-	DBusMessage *msg;
-	DBusPendingCall *pending;
-
-	msg = dbus_message_new_method_call("org.bluez", pr->adapter_path,
-			"org.bluez.Adapter", "GetRemoteServiceRecord");
-	if (!msg)
-		return -1;
-
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &pr->addr,
-			DBUS_TYPE_UINT32, &handle,
-			DBUS_TYPE_INVALID);
-
-	if (dbus_connection_send_with_reply(pr->conn, msg, &pending, -1) == FALSE) {
-		error("Can't send D-Bus message.");
-		dbus_message_unref(msg);
-		return -1;
-	}
-
-	dbus_pending_call_set_notify(pending, cb, pr, NULL);
-	dbus_message_unref(msg);
-	dbus_pending_call_unref(pending);
-
-	return 0;
-}
-
-static void pan_handle_reply(DBusPendingCall *call, void *data)
-{
-	struct pending_reply *pr = data;
-	DBusMessage *reply = dbus_pending_call_steal_reply(call);
-	DBusError derr;
-	uint32_t *phandle;
-	int len;
-
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		if (dbus_error_has_name(&derr,
-				"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connection_failed(pr->conn, pr->msg, derr.message);
-		else
-			err_not_supported(pr->conn, pr->msg);
-
-		error("GetRemoteServiceHandles: %s(%s)", derr.name,
-				derr.message);
-		goto fail;
-	}
-
-	if (!dbus_message_get_args(reply, &derr,
-				DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32, &phandle,
-				&len, DBUS_TYPE_INVALID)) {
-		err_not_supported(pr->conn, pr->msg);
-		error("%s: %s", derr.name, derr.message);
-		goto fail;
-	}
-
-	if (!len) {
-		err_not_supported(pr->conn, pr->msg);
-		goto fail;
-	}
-
-	if (get_record(pr, *phandle, pan_record_reply) < 0) {
-		err_not_supported(pr->conn, pr->msg);
-		goto fail;
-	}
-
-	dbus_message_unref(reply);
-	return;
-fail:
-	dbus_error_free(&derr);
+	g_free(desc);
 	pending_reply_free(pr);
 }
 
-static int get_handles(struct pending_reply *pr,
-			DBusPendingCallNotifyFunction cb)
-{
-	DBusMessage *msg;
-	DBusPendingCall *pending;
-	const char *uuid;
-
-	msg = dbus_message_new_method_call("org.bluez", pr->adapter_path,
-			"org.bluez.Adapter", "GetRemoteServiceHandles");
-	if (!msg)
-		return -1;
-
-	uuid = bnep_uuid(pr->id);
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &pr->addr,
-			DBUS_TYPE_STRING, &uuid,
-			DBUS_TYPE_INVALID);
-
-	if (dbus_connection_send_with_reply(pr->conn, msg, &pending, -1) == FALSE) {
-		error("Can't send D-Bus message.");
-		dbus_message_unref(msg);
-		return -1;
-	}
-
-	dbus_pending_call_set_notify(pending, cb, pr, NULL);
-	dbus_message_unref(msg);
-	dbus_pending_call_unref(pending);
-
-	return 0;
-}
-
-static DBusHandlerResult list_servers(DBusConnection *conn, DBusMessage *msg,
-					void *data)
+static DBusMessage *list_servers(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
 	return list_paths(conn, msg, server_paths);
 }
 
-static DBusHandlerResult find_server(DBusConnection *conn,
-						DBusMessage *msg, void *data)
+static DBusMessage *find_server(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
-	DBusError derr;
 	const char *pattern;
 	const char *path;
 	GSList *list;
-	DBusMessage *reply;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &pattern,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
+						DBUS_TYPE_INVALID) == FALSE)
+		return NULL;
 
 	for (list = server_paths; list; list = list->next) {
 		path = (const char *) list->data;
@@ -431,29 +305,21 @@ static DBusHandlerResult find_server(DBusConnection *conn,
 			break;
 	}
 
-	if (list == NULL) {
-		err_failed(conn, msg, "No such server");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (list == NULL)
+		return does_not_exist(msg);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult list_connections(DBusConnection *conn,
+static DBusMessage *list_connections(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	return list_paths(conn, msg, connection_paths);
 }
 
-static GSList * find_connection_pattern(DBusConnection *conn,
-					const char *pattern)
+static GSList *find_connection_pattern(DBusConnection *conn,
+						const char *pattern)
 {
 	const char *path;
 	GSList *list;
@@ -470,87 +336,70 @@ static GSList * find_connection_pattern(DBusConnection *conn,
 	return list;
 }
 
-static DBusHandlerResult find_connection(DBusConnection *conn,
+static DBusMessage *find_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
-	DBusError derr;
 	const char *pattern;
 	const char *path;
 	GSList *list;
-	DBusMessage *reply;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &pattern,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
+						DBUS_TYPE_INVALID) == FALSE)
+		return NULL;
 
 	list = find_connection_pattern(conn, pattern);
-
-	if (list == NULL) {
-		err_failed(conn, msg, "No such connection");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (list == NULL)
+		return does_not_exist(msg);
 
 	path = list->data;
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult create_connection(DBusConnection *conn,
+static DBusMessage *create_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct pending_reply *pr;
-	DBusError derr;
 	const char *addr;
 	const char *str;
 	bdaddr_t src;
 	uint16_t id;
-	int dev_id;
+	int dev_id, err;
 	char key[32];
 	GSList *l;
+	uuid_t uuid;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &addr,
-				DBUS_TYPE_STRING, &str,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &addr,
+			DBUS_TYPE_STRING, &str, DBUS_TYPE_INVALID) == FALSE)
+		return NULL;
 
 	id = bnep_service_id(str);
 	if (id != BNEP_SVC_GN && id != BNEP_SVC_NAP && id != BNEP_SVC_PANU)
-		return err_invalid_args(conn, msg, "Not supported");
+		return not_supported(msg);
 
 	snprintf(key, 32, "%s#%s", addr, bnep_name(id));
 
 	/* Checks if the connection was already been made */
 	for (l = connection_paths; l; l = l->next) {
-		if (connection_find_data(l->data, key) == 0) {
-			err_already_exists(conn, msg,
-						"Connection Already exists");
-			return DBUS_HANDLER_RESULT_HANDLED;
-		}
+		if (connection_find_data(l->data, key) == 0)
+			return already_exists(msg);
 	}
 
 	bacpy(&src, BDADDR_ANY);
-	dev_id = hci_get_route(NULL);
-	if ((dev_id < 0) ||  (hci_devba(dev_id, &src) < 0))
-		return err_failed(conn, msg, "Adapter not available");
+	dev_id = hci_get_route(&src);
+	if (dev_id < 0 || hci_devba(dev_id, &src) < 0)
+		return adapter_not_available(msg);
 
 	pr = g_new0(struct pending_reply, 1);
+
+	/* FIXME just to maintain compatibility */
+	pr->adapter_path = g_strdup_printf("/org/bluez/hci%d", dev_id);
+	if (!pr->adapter_path) {
+		pending_reply_free (pr);
+		return adapter_not_available(msg);
+	}
+
 	pr->conn = dbus_connection_ref(conn);
 	pr->msg = dbus_message_ref(msg);
 	bacpy(&pr->src, &src);
@@ -559,58 +408,48 @@ static DBusHandlerResult create_connection(DBusConnection *conn,
 	pr->id = id;
 	pr->path = g_new0(char, MAX_PATH_LENGTH);
 	snprintf(pr->path, MAX_PATH_LENGTH,
-			NETWORK_PATH"/connection%d", net_uid++);
+				NETWORK_PATH "/connection%d", net_uid++);
 
-	pr->adapter_path = g_malloc0(16);
-	snprintf(pr->adapter_path, 16, "/org/bluez/hci%d", dev_id);
+	sdp_uuid16_create(&uuid, pr->id);
+	err = bt_search_service(&pr->src, &pr->dst, &uuid, records_cb, pr,
+				NULL);
+	if (err < 0) {
+		pending_reply_free(pr);
+		return not_supported(msg);
+	}
 
-	if (get_handles(pr, pan_handle_reply) < 0)
-		return err_not_supported(conn, msg);
-
-	return DBUS_HANDLER_RESULT_HANDLED;
+	return NULL;
 }
 
-static DBusHandlerResult remove_connection(DBusConnection *conn,
+static DBusMessage *remove_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	return remove_path(conn, msg, &connection_paths, "ConnectionRemoved");
 }
 
-static DBusHandlerResult last_connection(DBusConnection *conn,
+static DBusMessage *last_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	const char *path;
-	DBusMessage *reply;
 
 	if (connection_paths == NULL ||
-		g_slist_length (connection_paths) == 0) {
-		err_does_not_exist(conn, msg, "No such connection");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+				g_slist_length(connection_paths) == 0)
+		return does_not_exist(msg);
 
 	path = last_connection_used(conn);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult default_connection(DBusConnection *conn,
+static DBusMessage *default_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	const char *path;
-	DBusMessage *reply;
 
 	if (connection_paths == NULL ||
-		g_slist_length (connection_paths) == 0) {
-		err_does_not_exist(conn, msg, "No such connection");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+				g_slist_length (connection_paths) == 0)
+		return does_not_exist(msg);
 
 	path = g_slist_nth_data (connection_paths, default_index);
 
@@ -619,87 +458,63 @@ static DBusHandlerResult default_connection(DBusConnection *conn,
 		connection_store(path, TRUE);
 	}
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult change_default_connection(DBusConnection *conn,
+static DBusMessage *change_default_connection(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	const char *path;
 	const char *pattern;
-	DBusMessage *reply;
-	DBusError derr;
 	GSList *list;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_STRING, &pattern,
-				DBUS_TYPE_INVALID)) {
-		err_invalid_args(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &pattern,
+						DBUS_TYPE_INVALID) == FALSE)
+		return NULL;
 
 	if (connection_paths == NULL ||
-		g_slist_length (connection_paths) == 0) {
-		err_does_not_exist(conn, msg, "No such connection");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+				g_slist_length(connection_paths) == 0)
+		return does_not_exist(msg);
 
-	list = g_slist_find_custom(connection_paths, pattern, (GCompareFunc) strcmp);
+	list = g_slist_find_custom(connection_paths, pattern,
+						(GCompareFunc) strcmp);
 
 	/* Find object path via pattern */
 	if (list == NULL) {
 		list = find_connection_pattern(conn, pattern);
+		if (list == NULL)
+			return does_not_exist(msg);
 
-		if (list == NULL) {
-			err_failed(conn, msg, "No such connection");
-			return DBUS_HANDLER_RESULT_HANDLED;
-		}
-		else
-			path = list->data;
-	}
-	else
+		path = list->data;
+	} else
 		path = list->data;
 
-	default_index = g_slist_position (connection_paths, list);
+	default_index = g_slist_position(connection_paths, list);
 	connection_store(path, TRUE);
 
-	dbus_connection_emit_signal(connection, NETWORK_PATH,
+	g_dbus_emit_signal(connection, NETWORK_PATH,
 					NETWORK_MANAGER_INTERFACE,
 					"DefaultConnectionChanged",
 					DBUS_TYPE_STRING, &path,
 					DBUS_TYPE_INVALID);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &path,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &path,
+							DBUS_TYPE_INVALID);
 }
 
-static void manager_unregister(DBusConnection *conn, void *data)
+static void manager_unregister(void *data)
 {
 	info("Unregistered manager path");
 
 	if (server_paths) {
-		g_slist_foreach(server_paths, (GFunc)g_free, NULL);
+		g_slist_foreach(server_paths, (GFunc) g_free, NULL);
 		g_slist_free(server_paths);
 		server_paths = NULL;
 	}
 
 	if (connection_paths) {
-		g_slist_foreach(connection_paths, (GFunc)g_free, NULL);
+		g_slist_foreach(connection_paths, (GFunc) g_free, NULL);
 		g_slist_free(connection_paths);
 		connection_paths = NULL;
 	}
@@ -743,7 +558,7 @@ static void parse_stored_connection(char *key, char *value, void *data)
 		return;
 
 	snprintf(path, MAX_PATH_LENGTH,
-			NETWORK_PATH"/connection%d", net_uid++);
+				NETWORK_PATH "/connection%d", net_uid++);
 
 	/* Parsing the value: name and description */
 	ptr = strchr(value, ':');
@@ -795,10 +610,9 @@ static void register_connections_stored(const char *adapter)
 	/* Check default connection for current default adapter */
 	bacpy(&default_src, BDADDR_ANY);
 	dev_id = hci_get_route(&default_src);
-	if (dev_id < 0)
+	if (dev_id < 0 || hci_devba(dev_id, &default_src) < 0)
 		return;
 
-	hci_devba(dev_id, &default_src);
 	if (bacmp(&default_src, &src) != 0)
 		return;
 
@@ -818,7 +632,10 @@ static void register_server(uint16_t id)
 	bdaddr_t src;
 	int dev_id;
 
-	snprintf(path, MAX_PATH_LENGTH, NETWORK_PATH"/%s", bnep_name(id));
+	if (!conf->server_enabled)
+		return;
+
+	snprintf(path, MAX_PATH_LENGTH, NETWORK_PATH "/%s", bnep_name(id));
 
 	if (g_slist_find_custom(server_paths, path,
 				(GCompareFunc) strcmp))
@@ -826,11 +643,9 @@ static void register_server(uint16_t id)
 
 	bacpy(&src, BDADDR_ANY);
 	dev_id = hci_get_route(&src);
-	if (dev_id < 0)
+	if (dev_id < 0 || hci_devba(dev_id, &src))
 		return;
 
-	hci_devba(dev_id, &src);
-
 	if (server_register(path, &src, id) < 0)
 		return;
 
@@ -859,8 +674,7 @@ static void register_servers_stored(const char *adapter, const char *profile)
 	str2ba(adapter, &src);
 
 	if (stat (filename, &s) == 0 && (s.st_mode & __S_IFREG)) {
-		snprintf(path, MAX_PATH_LENGTH,
-			NETWORK_PATH"/%s", profile);
+		snprintf(path, MAX_PATH_LENGTH, NETWORK_PATH "/%s", profile);
 		if (server_register_from_file(path, &src, id, filename) == 0) {
 			server_paths = g_slist_append(server_paths,
 						g_strdup(path));
@@ -885,47 +699,81 @@ static void register_stored(void)
 			continue;
 
 		/* Connection objects */
-		register_connections_stored(de->d_name);
+		if (conf->connection_enabled)
+			register_connections_stored(de->d_name);
 
-		/* NAP objects */
-		register_servers_stored(de->d_name, "nap");
+		/* Server objects */
+		if (conf->server_enabled) {
+			/* NAP objects */
+			register_servers_stored(de->d_name, "nap");
 
-		/* GN objects */
-		register_servers_stored(de->d_name, "gn");
+			/* GN objects */
+			register_servers_stored(de->d_name, "gn");
 
-		/* PANU objects */
-		register_servers_stored(de->d_name, "panu");
+			/* PANU objects */
+			register_servers_stored(de->d_name, "panu");
+		}
 	}
 
 	closedir(dir);
 }
 
-static DBusMethodVTable manager_methods[] = {
-	{ "ListServers",	list_servers,		"",	"as"	},
-	{ "FindServer",		find_server,		"s",	"s"	},
-	{ "ListConnections",	list_connections,	"",	"as"	},
-	{ "FindConnection",	find_connection,	"s",	"s"	},
-	{ "CreateConnection",	create_connection,	"ss",	"s"	},
-	{ "RemoveConnection",	remove_connection,	"s",	""	},
-	{ "LastConnection",	last_connection,	"",	"s"	},
-	{ "DefaultConnection",	default_connection,	"",	"s"	},
-	{ "ChangeDefaultConnection", change_default_connection, "s", "s"},
-	{ NULL, NULL, NULL, NULL }
+static GDBusMethodTable connection_methods[] = {
+	{ "ListConnections",	"",	"as",	list_connections	},
+	{ "FindConnection",	"s",	"s",	find_connection		},
+	{ "CreateConnection",	"ss",	"s",	create_connection,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "RemoveConnection",	"s",	"",	remove_connection	},
+	{ "LastConnection",	"",	"s",	last_connection		},
+	{ "DefaultConnection",	"",	"s",	default_connection	},
+	{ "ChangeDefaultConnection", "s", "s", change_default_connection },
+	{ }
 };
 
-static DBusSignalVTable manager_signals[] = {
-	{ "ConnectionCreated",	"s"	},
-	{ "ConnectionRemoved",	"s"	},
-	{ "DefaultConnectionChanged", "s" },
-	{ NULL, NULL }
+static GDBusSignalTable connection_signals[] = {
+	{ "ConnectionCreated",		"s"	},
+	{ "ConnectionRemoved",		"s"	},
+	{ "DefaultConnectionChanged",	"s"	},
+	{ }
 };
 
-int network_init(DBusConnection *conn, struct network_conf *service_conf)
+static GDBusMethodTable server_methods[] = {
+	{ "ListServers",	"",	"as",	list_servers	},
+	{ "FindServer",		"s",	"s",	find_server	},
+	{ }
+};
+
+static GDBusMethodTable manager_methods[] = {
+	{ "ListServers",	"",	"as",	list_servers	},
+	{ "FindServer",		"s",	"s",	find_server	},
+	{ "ListConnections",	"",	"as",	list_connections	},
+	{ "FindConnection",	"s",	"s",	find_connection		},
+	{ "CreateConnection",	"ss",	"s",	create_connection,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "RemoveConnection",	"s",	"",	remove_connection	},
+	{ "LastConnection",	"",	"s",	last_connection		},
+	{ "DefaultConnection",	"",	"s",	default_connection	},
+	{ "ChangeDefaultConnection", "s", "s", change_default_connection },
+	{ }
+};
+
+int network_manager_init(DBusConnection *conn, struct network_conf *service_conf)
 {
+	GDBusMethodTable *methods = NULL;
+	GDBusSignalTable *signals = NULL;
+
 	conf = service_conf;
 
-	if (bridge_init(conf->gn_iface, conf->nap_iface) < 0) {
-		error("Can't init bridge module");
+	if (conf->server_enabled && conf->connection_enabled) {
+		methods = manager_methods;
+		signals = connection_signals;
+	} else if (conf->connection_enabled) {
+		methods = connection_methods;
+		signals = connection_signals;
+	} else if (conf->server_enabled)
+		methods = server_methods;
+	else {
+		error ("All interfaces were disabled");
 		return -1;
 	}
 
@@ -940,34 +788,30 @@ int network_init(DBusConnection *conn, struct network_conf *service_conf)
 	 * (setup connection request) contains the destination service
 	 * field that defines which service the source is connecting to.
 	 */
-	if (server_init(conn, conf->iface_prefix, conf->security) < 0)
-		return -1;
+	if (conf->server_enabled) {
+		if (bridge_init(conf->gn_iface, conf->nap_iface) < 0) {
+			error("Can't init bridge module");
+			return -1;
+		}
 
-	if (connection_init(conn, conf->iface_prefix) < 0)
- 		return -1;
-
-	if (!dbus_connection_create_object_path(conn, NETWORK_PATH,
-						NULL, manager_unregister)) {
-		error("D-Bus failed to create %s path", NETWORK_PATH);
-		return -1;
+		if (server_init(conn, conf->iface_prefix, conf->security) < 0)
+			return -1;
 	}
 
-	if (!dbus_connection_register_interface(conn, NETWORK_PATH,
-						NETWORK_MANAGER_INTERFACE,
-						manager_methods,
-						manager_signals, NULL)) {
+	if (conf->connection_enabled) {
+		if (connection_init(conn, conf->iface_prefix) < 0)
+			return -1;
+	}
+
+	if (g_dbus_register_interface(conn, NETWORK_PATH,
+					NETWORK_MANAGER_INTERFACE,
+					methods, signals, NULL,
+					NULL, manager_unregister) == FALSE) {
 		error("Failed to register %s interface to %s",
 				NETWORK_MANAGER_INTERFACE, NETWORK_PATH);
-		dbus_connection_destroy_object_path(connection, NETWORK_PATH);
 		return -1;
 	}
 
-	if (bridge_create(BNEP_SVC_GN) < 0)
-		error("Can't create GN bridge");
-
-	if (bridge_create(BNEP_SVC_NAP) < 0)
-		error("Can't create NAP bridge");
-
 	connection = dbus_connection_ref(conn);
 
 	info("Registered manager path:%s", NETWORK_PATH);
@@ -982,21 +826,20 @@ int network_init(DBusConnection *conn, struct network_conf *service_conf)
 	return 0;
 }
 
-void network_exit(void)
+void network_manager_exit(void)
 {
-	server_exit();
-	connection_exit();
-	dbus_connection_destroy_object_path(connection, NETWORK_PATH);
+	if (conf->server_enabled)
+		server_exit();
+
+	if (conf->connection_enabled)
+		connection_exit();
+
+	g_dbus_unregister_interface(connection, NETWORK_PATH,
+						NETWORK_MANAGER_INTERFACE);
 
 	dbus_connection_unref(connection);
 	connection = NULL;
 
-	if (bridge_remove(BNEP_SVC_GN) < 0)
-		error("Can't remove GN bridge");
-
-	if (bridge_remove(BNEP_SVC_NAP) < 0)
-		error("Can't remove NAP bridge");
-
 	bnep_cleanup();
 	bridge_cleanup();
 }
