@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2002-2006  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -30,7 +30,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <malloc.h>
+#include <string.h>
 #include <getopt.h>
 #include <signal.h>
 #include <sys/poll.h>
